import re


class RegexBuilder:
    """
    Notes:
        The way to have multiple negative look aheads/behinds is to change them together SO12689046

    References:
        .. [SO12689046] https://stackoverflow.com/questions/12689046/multiple-negative-lookbehind-assertions-in-python-regex
    """
    common_patterns = [
        {'key': 'word',      'pattern': r'\w', 'docs': r'An alphanumeric word'},
        {'key': 'non-word',  'pattern': r'\W', 'docs': r'Anything not a word'},
        {'key': 'space',     'pattern': r'\s', 'docs': r'Any space character including: " " "\t", "\n", "\r"'},
        {'key': 'non-space', 'pattern': r'\S', 'docs': r'Any non-space character'},
        {'key': 'digit',     'pattern': r'\d', 'docs': r'any number 0-9'},
        {'key': 'digit',     'pattern': r'\D', 'docs': r'any non-digit'},
        {'key': 'zero_or_more', 'pattern': r'*', 'docs': r'zero or more of the pattern to the left', 'alias': ['kleene_star']},
    ]

    def __init__(self):
        raise Exception('Use coerce instead')

    def lookahead(self, pat, positive=True):
        """
        A lookahead pattern that can be positive or negative

        looklook
        """
        if positive:
            return self.constructs['positive_lookahead'].format(pat=pat)
        else:
            return self.constructs['negative_lookahead'].format(pat=pat)

    def lookbehind(self, pat, positive=True):
        """
        A lookbehind pattern that can be positive or negative
        """
        if positive:
            return self.constructs['positive_lookbehind'].format(pat=pat)
        else:
            return self.constructs['negative_lookbehind'].format(pat=pat)

    def named_field(self, pat, name=None):
        if name is None:
            return self.constructs['group'].format(pat=pat)
        else:
            return self.constructs['named_field'].format(pat=pat, name=name)

    def bref_field(self, name):
        return self.constructs['backref_field'].format(name=name)

    def escape(self, pat):
        return re.escape(pat)

    def optional(self, pat):
        return r'{pat}?'.format(pat=pat)

    def group(self, pat):
        return self.constructs['group'].format(pat=pat)

    def oneof(self, *paterns):
        return self.group('|'.join(paterns))

    @classmethod
    def coerce(cls, backend='python'):
        if backend == 'python':
            cls = PythonRegexBuilder
        elif backend == 'vim':
            cls = VimRegexBuilder
        else:
            raise KeyError(backend)
        self = cls()
        return self

    @property
    def whitespace(self):
        return self.special['space'] + '*'

    @property
    def nongreedy(self):
        return self.special['nongreedy_kleene_star']

    @property
    def number(self):
        """
        Can match a generic floating point number

        References:
            https://www.regular-expressions.info/floatingpoint.html

        Example:
            >>> from xdev.regex_builder import *  # NOQA
            >>> b = PythonRegexBuilder()
            >>> pat = re.compile('^' + b.number + '$')
            >>> assert pat.match('3.4')
            >>> assert pat.match('3.4e-1')
            >>> assert pat.match('3.4')
            >>> assert pat.match('3.4e+1')
            >>> assert not pat.match('3.4a+1')

            >>> b = PythonRegexBuilder()
            >>> num_part = b.named_field(b.number, name='number')
            >>> space_part = b.named_field(' *', name='spaces')
            >>> unit_part = b.named_field('.*', name='unit')
            >>> pat = re.compile('^' + num_part + space_part + unit_part + '$')
            >>> pat.match('3.4').groupdict()
            >>> pat.match('3.1415 foobars').groupdict()
            >>> pat.match('3.1415foobars').groupdict()
            >>> pat.match('+3.1415e9foobars').groupdict()
        """
        exponent_part = '[eE][-+]?[0-9]+'
        decimal_part = r'[-+]?[0-9]*\.?[0-9]+'
        exponent_group = self.constructs['group'].format(pat=exponent_part)
        number_pat = decimal_part + exponent_group + '?'
        return number_pat


class VimRegexBuilder(RegexBuilder):
    """
    https://dev.to/iggredible/learning-vim-regex-26ep
    """

    vim_patterns = [
        {'key': 'nongreedy_zero_or_more', 'pattern': r'\{-}', 'docs': r'non-greedily matches zero or more of the pattern to the left', 'alias': ['nongreedy_kleene_star']},
    ]
    def __init__(self):
        self.constructs = {}
        self.constructs['positive_lookahead'] = r'\({pat}\)\@='
        self.constructs['negative_lookahead'] = r'\({pat}\)\@!'
        self.constructs['positive_lookbehind'] = r'\({pat}\)\@<='
        self.constructs['negative_lookbehind'] = r'\({pat}\)\@<!'
        self.constructs['word'] = r'\<{pat}\>'
        self.constructs['group'] = r'\({pat}\)'

        self.special = {}
        for item in self.common_patterns + self.vim_patterns:
            self.special[item['key']] = item['pattern']


class PythonRegexBuilder(RegexBuilder):
    r"""
    Contains helper methods to construct a regex

    Example:
        >>> b = PythonRegexBuilder()
        >>> pat_text = b.lookbehind('_') + r'v\d+' + b.optional(b.lookahead('_'))
        >>> pat = re.compile(pat_text)
        >>> print(pat.search('_v321_').group())
        v321
        >>> print(pat.search('_v321').group())
        v321
        >>> print(pat.search('fdsfds_v321_fdsfsd').group())
        v321
        >>> print(pat.search('fdsfds_v321fdsfsd').group())
        v321
        >>> print(pat.search('fdsfdsv321fdsfsd'))
        None

    Example:
        >>> # Test multiple negative lookbehind
        >>> b = PythonRegexBuilder()
        >>> suffix = 'foo'
        >>> neg_prefix1 = b.lookbehind('abc', positive=0)
        >>> neg_prefix2 = b.lookbehind('efg', positive=0)
        >>> pat1 = re.compile(neg_prefix1 + suffix)
        >>> pat2 = re.compile(neg_prefix2 + suffix)
        >>> patB = re.compile(neg_prefix1 + neg_prefix2 + suffix)
        >>> cases = ['abcfoo', 'efgfoo', 'hijfoo', 'foo']
        >>> print([bool(pat1.search(c)) for c in cases])
        >>> print([bool(pat2.search(c)) for c in cases])
        >>> print([bool(patB.search(c)) for c in cases])
        [False, True, True, True]
        [True, False, True, True]
        [False, False, True, True]

    References:
        https://www.dataquest.io/blog/regex-cheatsheet/
    """

    python_patterns = [
        {'key': 'nongreedy_zero_or_more', 'pattern': r'*?', 'docs': r'non-greedily matches zero or more of the pattern to the left', 'alias': ['nongreedy_kleene_star']},
        {'key': 'boundary', 'pattern': r'\b', 'docs': r'The boundary at the start or end of a word'},
        {'key': 'non-boundary', 'pattern': r'\B'},
        {'key': 'left-expr', 'pattern': r'\A'},
        {'key': 'right-expr', 'pattern': r'\Z'},
    ]
    def __init__(self):
        self.constructs = {}
        self.constructs['positive_lookahead'] = r'(?={pat})'
        self.constructs['negative_lookahead'] = r'(?!{pat})'
        self.constructs['positive_lookbehind'] = r'(?<={pat})'
        self.constructs['negative_lookbehind'] = r'(?<!{pat})'
        self.constructs['word'] = r'\b{pat}\b'
        self.constructs['group'] = r'({pat})'
        self.constructs['named_field'] = r'(?P<{name}>{pat})'
        self.constructs['backref_field'] = r'\g<{name}>'

        self.special = {}
        for item in self.common_patterns + self.python_patterns:
            self.special[item['key']] = item['pattern']



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import decimal
from collections import defaultdict
import ubelt as ub


def edit_distance(string1, string2):
    """
    Edit distance algorithm. String1 and string2 can be either
    strings or lists of strings

    Args:
        string1 (str | List[str]):
        string2 (str | List[str]):

    Requirements:
        pip install python-Levenshtein

    Example:
        >>> # xdoctest: +REQUIRES(module:Levenshtein)
        >>> string1 = 'hello world'
        >>> string2 = ['goodbye world', 'rofl', 'hello', 'world', 'lowo']
        >>> edit_distance(['hello', 'one'], ['goodbye', 'two'])
        >>> edit_distance('hello', ['goodbye', 'two'])
        >>> edit_distance(['hello', 'one'], 'goodbye')
        >>> edit_distance('hello', 'goodbye')
        >>> distmat = edit_distance(string1, string2)
        >>> result = ('distmat = %s' % (ub.repr2(distmat),))
        >>> print(result)
        >>> [7, 9, 6, 6, 7]
    """

    import Levenshtein
    isiter1 = ub.iterable(string1)
    isiter2 = ub.iterable(string2)
    strs1 = string1 if isiter1 else [string1]
    strs2 = string2 if isiter2 else [string2]
    distmat = [
        [Levenshtein.distance(str1, str2) for str2 in strs2]
        for str1 in strs1
    ]
    # broadcast
    if not isiter2:
        distmat = [row[0] for row in distmat]
    if not isiter1:
        distmat = distmat[0]
    return distmat


def knapsack(items, maxweight, method='iterative'):
    r"""
    Solve the knapsack problem by finding the most valuable subsequence of
    `items` subject that weighs no more than `maxweight`.

    Args:
        items (tuple): is a sequence of tuples `(value, weight, id_)`, where
            `value` is a number and `weight` is a non-negative integer, and
            `id_` is an item identifier.

        maxweight (numbers.Real):
            a non-negative number indicating the maximum weight of items we can
            take.

    Returns:
        tuple: (total_value, items_subset) - a pair whose first element is the
            sum of values in the most valuable subsequence, and whose second
            element is the subsequence. Subset may be different depending on
            implementation (ie top-odwn recusrive vs bottom-up iterative)

    References:
        http://codereview.stackexchange.com/questions/20569/dynamic-programming-solution-to-knapsack-problem
        http://stackoverflow.com/questions/141779/solving-the-np-complete-problem-in-xkcd
        http://www.es.ele.tue.nl/education/5MC10/Solutions/knapsack.pdf

    Example:
        >>> # ENABLE_DOCTEST
        >>> import ubelt as ub
        >>> items = [(4, 12, 0), (2, 1, 1), (6, 4, 2), (1, 1, 3), (2, 2, 4)]
        >>> maxweight = 15
        >>> total_value1, items_subset1 = knapsack(items, maxweight, method='iterative')
        >>> print('items_subset1 = {}'.format(ub.repr2(items_subset1, nl=1)))
        >>> print('total_value1 = {}'.format(ub.repr2(total_value1, nl=1)))

    Example:
        >>> # ENABLE_DOCTEST
        >>> # Solve https://xkcd.com/287/
        >>> weights = [2.15, 2.75, 3.35, 3.55, 4.2, 5.8] * 2
        >>> items = [(w, w, i) for i, w in enumerate(weights)]
        >>> maxweight = 15.05
        >>> total_value1, items_subset1 = knapsack(items, maxweight, method='iterative')
        >>> total_weight = sum([t[1] for t in items_subset1])
        >>> print('total_weight = %r' % (total_weight,))
        >>> print('items_subset1 = %r' % (items_subset1,))
        >>> #assert items_subset1 == items_subset, 'NOT EQ\n%r !=\n%r' % (items_subset1, items_subset)

    Timeit:
        # >>> import ubelt as ub
        # >>> setup = ub.codeblock(
        # >>>     '''
        #         import ubelt as ut
        #         weights = [215, 275, 335, 355, 42, 58] * 40
        #         items = [(w, w, i) for i, w in enumerate(weights)]
        #         maxweight = 2505
        #         #import numba
        #         #knapsack_numba = numba.autojit(knapsack_iterative)
        #         #knapsack_numba = numba.autojit(knapsack_iterative_numpy)
        #         ''')
        # >>> # Test load time
        # >>> stmt_list1 = ub.codeblock(
        # >>>     '''
        #         knapsack_iterative(items, maxweight)
        #         knapsack_ilp(items, maxweight)
        #         #knapsack_numba(items, maxweight)
        #         #knapsack_iterative_numpy(items, maxweight)
        #         ''').split('\n')

        # ut.util_dev.timeit_compare(stmt_list1, setup, int(5))

        from xdev.algo import *  # NOQA
        def knapsack_inputs(n):
            rng = np.random
            items = [
                (rng.randint(0, 10), rng.randint(0, 10), idx) for idx in range(n)]
            return {'items':  items, 'maxweight': 100}

        def input_wrapper(func):
            import functools
            @functools.wraps(func)
            def _wrap(kwargs):
                return func(**kwargs)
            return _wrap

        import perfplot
        perfplot.show(
            setup=knapsack_inputs,
            kernels=list(map(input_wrapper, [knapsack_iterative, knapsack_ilp])),
            n_range=[2 ** k for k in range(15)],
            xlabel="maxweight",
            equality_check=None,
        )

    """
    if method == 'iterative':
        return knapsack_iterative(items, maxweight)
    elif method == 'ilp':
        return knapsack_ilp(items, maxweight)
    else:
        raise NotImplementedError('[util_alg] knapsack method=%r' % (method,))
        #return knapsack_iterative_numpy(items, maxweight)


def knapsack_ilp(items, maxweight, verbose=False):
    """
    solves knapsack using an integer linear program

    Example:
        >>> # DISABLE_DOCTEST
        >>> import ubelt as ub
        >>> # Solve https://xkcd.com/287/
        >>> weights = [2.15, 2.75, 3.35, 3.55, 4.2, 5.8, 6.55]
        >>> values  = [2.15, 2.75, 3.35, 3.55, 4.2, 5.8, 6.55]
        >>> indices = ['mixed fruit', 'french fries', 'side salad',
        >>>            'hot wings', 'mozzarella sticks', 'sampler plate',
        >>>            'barbecue']
        >>> items = [(v, w, i) for v, w, i in zip(values, weights, indices)]
        >>> #items += [(3.95, 3.95, 'mystery plate')]
        >>> maxweight = 15.05
        >>> verbose = True
        >>> total_value, items_subset = knapsack_ilp(items, maxweight, verbose)
        >>> print('items_subset = %s' % (ub.repr2(items_subset, nl=1),))
    """
    import pulp
    # Given Input
    values  = [t[0] for t in items]
    weights = [t[1] for t in items]
    indices = [t[2] for t in items]
    # Formulate integer program
    prob = pulp.LpProblem("Knapsack", pulp.LpMaximize)
    # Solution variables
    x = pulp.LpVariable.dicts(name='x', indexs=indices,
                              lowBound=0, upBound=1, cat=pulp.LpInteger)
    # maximize objective function
    prob.objective = sum(v * x[i] for v, i in zip(values, indices))
    # subject to
    prob.add(sum(w * x[i] for w, i in zip(weights, indices)) <= maxweight)
    # Solve using with solver like CPLEX, GLPK, or SCIP.
    #pulp.CPLEX().solve(prob)
    pulp.PULP_CBC_CMD().solve(prob)
    # Read solution
    flags = [x[i].varValue for i in indices]
    total_value = sum([val for val, flag in zip(values, flags) if flag])
    items_subset = [item for item, flag in zip(items, flags) if flag]
    # Print summary
    if verbose:
        print(prob)
        print('OPT:')
        print('\n'.join(['    %s = %s' % (x[i].name, x[i].varValue) for i in indices]))
        print('total_value = %r' % (total_value,))
    return total_value, items_subset


def number_of_decimals(num):
    r"""
    Args:
        num (float):

    References:
        stackoverflow.com/questions/6189956/finding-decimal-places

    Example:
        >>> # ENABLE_DOCTEST
        >>> num = 15.05
        >>> result = number_of_decimals(num)
        >>> print(result)
        2
    """
    exp = decimal.Decimal(str(num)).as_tuple().exponent
    return max(0, -exp)


def knapsack_iterative(items, maxweight):
    # Knapsack requires integral weights
    weights = [t[1] for t in items]
    max_exp = max([number_of_decimals(w_) for w_ in weights])
    coeff = 10 ** max_exp
    # Adjust weights to be integral
    int_maxweight = int(maxweight * coeff)
    int_items = [(v, int(w * coeff), idx) for v, w, idx in items]
    """
    items = int_items
    maxweight = int_maxweight
    """
    return knapsack_iterative_int(int_items, int_maxweight)


def knapsack_iterative_int(items, maxweight):
    r"""
    Iterative knapsack method

    Math:
        maximize \sum_{i \in T} v_i
        subject to \sum_{i \in T} w_i \leq W

    Notes:
        dpmat is the dynamic programming memoization matrix.
        dpmat[i, w] is the total value of the items with weight at most W
        T is idx_subset, the set of indicies in the optimal solution

    Example:
        >>> # ENABLE_DOCTEST
        >>> weights = [1, 3, 3, 5, 2, 1] * 2
        >>> items = [(w, w, i) for i, w in enumerate(weights)]
        >>> maxweight = 10
        >>> items = [(.8, 700, 0)]
        >>> maxweight = 2000
        >>> print('maxweight = %r' % (maxweight,))
        >>> print('items = %r' % (items,))
        >>> total_value, items_subset = knapsack_iterative_int(items, maxweight)
        >>> total_weight = sum([t[1] for t in items_subset])
        >>> print('total_weight = %r' % (total_weight,))
        >>> print('items_subset = %r' % (items_subset,))
        >>> result =  'total_value = %.2f' % (total_value,)
        >>> print(result)
        total_value = 0.80

    Ignore:
        DPMAT = [[dpmat[r][c] for c in range(maxweight)] for r in range(len(items))]
        KMAT  = [[kmat[r][c] for c in range(maxweight)] for r in range(len(items))]
    """
    values  = [t[0] for t in items]
    weights = [t[1] for t in items]
    maxsize = maxweight + 1
    # Sparse representation seems better
    dpmat = defaultdict(lambda: defaultdict(lambda: np.inf))
    kmat = defaultdict(lambda: defaultdict(lambda: False))
    idx_subset = []  # NOQA
    for w in range(maxsize):
        dpmat[0][w] = 0
    # For each item consider to include it or not
    for idx in range(len(items)):
        item_val = values[idx]
        item_weight = weights[idx]
        # consider at each possible bag size
        for w in range(maxsize):
            valid_item = item_weight <= w
            if idx > 0:
                prev_val = dpmat[idx - 1][w]
                prev_noitem_val = dpmat[idx - 1][w - item_weight]
            else:
                prev_val = 0
                prev_noitem_val = 0
            withitem_val = item_val + prev_noitem_val
            more_valuable = withitem_val > prev_val
            if valid_item and more_valuable:
                dpmat[idx][w] = withitem_val
                kmat[idx][w] = True
            else:
                dpmat[idx][w] = prev_val
                kmat[idx][w] = False
    # Trace backwards to get the items used in the solution
    K = maxweight
    for idx in reversed(range(len(items))):
        if kmat[idx][K]:
            idx_subset.append(idx)
            K = K - weights[idx]
    idx_subset = sorted(idx_subset)
    items_subset = [items[i] for i in idx_subset]
    total_value = dpmat[len(items) - 1][maxweight]
    return total_value, items_subset


def knapsack_iterative_numpy(items, maxweight):
    r"""
    Iterative knapsack method

    maximize \sum_{i \in T} v_i
    subject to \sum_{i \in T} w_i \leq W

    Notes:
        dpmat is the dynamic programming memoization matrix.
        dpmat[i, w] is the total value of the items with weight at most W
        T is the set of indicies in the optimal solution
    """
    #import numpy as np
    items = np.array(items)
    weights = items.T[1]
    # Find maximum decimal place (this problem is in NP)
    max_exp = max([number_of_decimals(w_) for w_ in weights])
    coeff = 10 ** max_exp
    # Adjust weights to be integral
    weights = (weights * coeff).astype(np.int)
    values  = items.T[0]
    MAXWEIGHT = int(maxweight * coeff)
    W_SIZE = MAXWEIGHT + 1

    dpmat = np.full((len(items), W_SIZE), np.inf)
    kmat = np.full((len(items), W_SIZE), 0, dtype=np.bool)
    idx_subset = []

    for w in range(W_SIZE):
        dpmat[0][w] = 0
    for idx in range(1, len(items)):
        item_val = values[idx]
        item_weight = weights[idx]
        for w in range(W_SIZE):
            valid_item = item_weight <= w
            prev_val = dpmat[idx - 1][w]
            if valid_item:
                prev_noitem_val = dpmat[idx - 1][w - item_weight]
                withitem_val = item_val + prev_noitem_val
                more_valuable = withitem_val > prev_val
            else:
                more_valuable = False
            dpmat[idx][w] = withitem_val if more_valuable else prev_val
            kmat[idx][w] = more_valuable
    K = MAXWEIGHT
    for idx in reversed(range(1, len(items))):
        if kmat[idx, K]:
            idx_subset.append(idx)
            K = K - weights[idx]
    idx_subset = sorted(idx_subset)
    items_subset = [items[i] for i in idx_subset]
    total_value = dpmat[len(items) - 1][MAXWEIGHT]
    return total_value, items_subset


#def knapsack_all_solns(items, maxweight):
#    """
#    TODO: return all optimal solutions to the knapsack problem

#    References:
#        stackoverflow.com/questions/30554290/all-solutions-from-knapsack-dp-matrix

#    >>> items = [(1, 2, 0), (1, 3, 1), (1, 4, 2), (1, 3, 3), (1, 3, 4), (1, 5, 5), (1, 4, 6), (1, 1, 7), (1, 1, 8), (1, 3, 9)]
#    >>> weights = ut.get_list_column(items, 1)
#    >>> maxweight = 6
#    """


def knapsack_greedy(items, maxweight):
    r"""
    non-optimal greedy version of knapsack algorithm
    does not sort input. Sort the input by largest value
    first if desired.

    Args:
        `items` (tuple): is a sequence of tuples `(value, weight, id_)`, where `value`
            is a scalar and `weight` is a non-negative integer, and `id_` is an
            item identifier.

        `maxweight` (scalar):  is a non-negative integer.

    Example:
        >>> # ENABLE_DOCTEST
        >>> items = [(4, 12, 0), (2, 1, 1), (6, 4, 2), (1, 1, 3), (2, 2, 4)]
        >>> maxweight = 15
        >>> total_value, items_subset = knapsack_greedy(items, maxweight)
        >>> result =  'total_value = %r\n' % (total_value,)
        >>> result += 'items_subset = %r' % (items_subset,)
        >>> print(result)
        total_value = 7
        items_subset = [(4, 12, 0), (2, 1, 1), (1, 1, 3)]
    """
    items_subset = []
    total_weight = 0
    total_value = 0
    for item in items:
        value, weight = item[0:2]
        if total_weight + weight > maxweight:
            continue
        else:
            items_subset.append(item)
            total_weight += weight
            total_value += value
    return total_value, items_subset


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/xdev/xdev/algo.py all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import sys
from functools import partial
from xdoctest.dynamic_analysis import get_parent_frame, get_stack_frame
from xdev import util
import time


def embed(parent_locals=None, parent_globals=None, exec_lines=None,
          remove_pyqt_hook=True, n=0):
    """
    Starts interactive session. Similar to keyboard command in matlab.
    Wrapper around IPython.embed.

    Note:
        Contains helper logic to allow the developer to more easilly exit the
        program if embed is called in a loop. Specifically if you want to quit
        and not embed again, then set qqq=1 before exiting the embed session.
    """
    from xdev._ipython_ext import embed2
    # import IPython
    import xdev  # NOQA
    import xdev as xd  # NOQA
    import os

    if parent_globals is None:
        parent_globals = get_parent_frame(n=n).f_globals
    if parent_locals is None:
        parent_locals = get_parent_frame(n=n).f_locals

    stackdepth = n  # NOQA
    getframe = partial(get_parent_frame, n=n)  # NOQA

    # exec(execstr_dict(parent_globals, 'parent_globals'))
    # exec(execstr_dict(parent_locals,  'parent_locals'))
    print('')
    print('================')
    print(util.bubbletext('EMBEDDING'))
    print('================')
    print('[xdev.embed] embedding')

    if os.environ.get('XDEV_USE_GUITOOL', ''):
        # I don't think this is needed anymore
        try:
            if remove_pyqt_hook:
                try:
                    import guitool
                    guitool.remove_pyqt_input_hook()
                except (ImportError, ValueError, AttributeError) as ex:
                    print('ex = {!r}'.format(ex))
                    pass
                # make qt not loop forever (I had qflag loop forever with this off)
        except ImportError as ex:
            print(ex)

    if 1:
        # Disable common annoyance loggers
        import logging
        logging.getLogger('parso').setLevel(logging.INFO)

    #from IPython.config.loader import Config
    # cfg = Config()
    #config_dict = {}
    #if exec_lines is not None:
    #    config_dict['exec_lines'] = exec_lines
    #IPython.embed(**config_dict)
    # print('[xdev.embed]  Get stack location with: ')
    # print('[xdev.embed] get_parent_frame(n=8).f_code.co_name')
    print('[xdev.embed] use xdev.fix_embed_globals() to address https://github.com/ipython/ipython/issues/62')
    print('[xdev.embed] set EXIT_NOW or qqq=1 to hard exit on unembed')
    #print('set iup to True to draw plottool stuff')
    # print('[util] call %pylab qt4 to get plottool stuff working')
    once = True
    # Allow user to set iup and redo the loop
    while once or vars().get('iup', False):
        if not once:
            # SUPER HACKY WAY OF GETTING FIGURES ON THE SCREEN BETWEEN UPDATES
            #vars()['iup'] = False
            # ALL YOU NEED TO DO IS %pylab qt4
            print('re-emebeding')
            #import plottool as pt
            #pt.update()
            #(pt.present())
            for _ in range(100):
                time.sleep(.01)

        once = False
        #vars().get('iup', False):
        print('[util] calling IPython.embed()')
        """
        Notes:
            /usr/local/lib/python2.7/dist-packages/IPython/terminal/embed.py
            IPython.terminal.embed.InteractiveShellEmbed

            # instance comes from  IPython.config.configurable.SingletonConfigurable.instance
        """
        #c = IPython.Config()
        #c.InteractiveShellApp.exec_lines = [
        #    '%pylab qt4',
        #    '%gui qt4',
        #    "print 'System Ready!'",
        #]
        #IPython.embed(config=c)
        parent_ns = parent_globals.copy()
        parent_ns.update(parent_locals)
        locals().update(parent_ns)
        try:
            embed2()
            # IPython.embed()
        except RuntimeError as ex:
            print('ex = {!r}'.format(ex))
            print('Failed to open ipython')
        #config = IPython.terminal.ipapp.load_default_config()
        #config.InteractiveShellEmbed = config.TerminalInteractiveShell
        #module = sys.modules[parent_globals['__name__']]
        #config['module'] = module
        #config['module'] = module
        #embed2(stack_depth=n + 2 + 1)
        #IPython.embed(config=config)
        #IPython.embed(config=config)
        #IPython.embed(module=module)
        # Exit python immediately if specifed
        if vars().get('EXIT_NOW', False) or vars().get('qqq', False):
            print('[xdev.embed] EXIT_NOW specified')
            sys.exit(1)


def embed_if_requested(n=0):
    """
    Calls xdev.embed conditionally based on the environment.

    Useful in cases where you want to leave the embed call around, but you dont
    want it to trigger in normal circumstances.

    Specifically, embed is only called if the environment variable XDEV_EMBED
    exists or if --xdev-embed is in sys.argv.
    """
    import os
    import ubelt as ub
    import xdev
    if os.environ.get('XDEV_EMBED', '') or ub.argflag('--xdev-embed'):
        xdev.embed(n=n + 1)


class EmbedOnException(object):
    """
    Context manager which embeds in ipython if an exception is thrown
    """
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __call__(self):
        return self

    def __exit__(__self, __type, __value, __trace):
        if __trace is not None:
            print('!!! EMBED ON EXCEPTION !!!')
            print('[util_dbg] %r in context manager!: %s ' % (__type, str(__value)))
            import traceback
            traceback.print_exception(__type, __value, __trace)
            # Grab the context of the frame where the failure occurred
            __trace_globals = __trace.tb_frame.f_globals
            __trace_locals = __trace.tb_frame.f_locals
            __trace_ns = __trace_globals.copy()
            __trace_ns.update(__trace_locals)
            # Hack to bring back names that we clobber
            if '__self' in __trace_ns:
                __self = __trace_ns['__self']
            locals().update(__trace_ns)  # I don't think this does anything
            embed()


def fix_embed_globals():
    """
    HACK adds current locals() to globals().
    Can be dangerous.

    References:
        https://github.com/ipython/ipython/issues/62

    Solves the following issue:

        def foo():
            x = 5
            # You embed here
            import xdev
            xdev.embed()

            '''
            Now you try and run this line manually but you get a NameError

            result = [x + i for i in range(10)]

            No problem, just use. It changes all local variables to globals
            xdev.fix_embed_globals()

            '''
            result = [x + i for i in range(10)]

        foo()
    """
    # Get the stack frame of whoever called this function
    frame = get_stack_frame(n=1)
    # Hack all of the local variables to be global variables
    frame.f_globals.update(frame.f_locals)
    # Leave some trace that we did this
    frame.f_globals['_did_xdev_fix_embed_globals'] = True


embed_on_exception_context = EmbedOnException()



import sys
from collections import defaultdict
import networkx as nx
from networkx.utils import open_file


### See: https://github.com/networkx/networkx/pull/5602


class _AsciiBaseGlyphs:
    empty = "+"
    newtree_last = "+-- "
    newtree_mid = "+-- "
    endof_forest = "    "
    within_forest = ":   "
    within_tree = "|   "


class AsciiDirectedGlyphs(_AsciiBaseGlyphs):
    last = "L-> "
    mid = "|-> "
    backedge = "<-"


class AsciiUndirectedGlyphs(_AsciiBaseGlyphs):
    last = "L-- "
    mid = "|-- "
    backedge = "-"


class _UtfBaseGlyphs:
    # Notes on available box and arrow characters
    # https://en.wikipedia.org/wiki/Box-drawing_character
    # https://stackoverflow.com/questions/2701192/triangle-arrow
    empty = "╙"
    newtree_last = "╙── "
    newtree_mid = "╟── "
    endof_forest = "    "
    within_forest = "╎   "
    within_tree = "│   "


class UtfDirectedGlyphs(_UtfBaseGlyphs):
    last = "└─╼ "
    mid = "├─╼ "
    backedge = "╾"


class UtfUndirectedGlyphs(_UtfBaseGlyphs):
    last = "└── "
    mid = "├── "
    backedge = "─"


def generate_network_text(
    graph, with_labels=True, sources=None, max_depth=None, ascii_only=False
):
    """Generate lines in the "network text" format

    This works via a depth-first traversal of the graph and writing a line for
    each unique node encountered. Non-tree edges are written to the right of
    each node, and connection to a non-tree edge is indicated with an ellipsis.
    This representation works best when the input graph is a forest, but any
    graph can be represented.

    This notation is original to networkx, although it is simple enough that it
    may be known in existing literature. See #5602 for details. The procedure
    is summarized as follows:

    1. Given a set of source nodes (which can be specified, or automatically
    discovered via finding the (strongly) connected components and choosing one
    node with minimum degree from each), we traverse the graph in depth first
    order.

    2. Each reachable node will be printed exactly once on it's own line.

    3. Edges are indicated in one of three ways:

        a. a parent "L-style" connection on the upper left. This corresponds to
        a traversal in the directed DFS tree.

        b. a backref "<-style" connection shown directly on the right. For
        directed graphs, these are drawn for any incoming edges to a node that
        is not a parent edge. For undirected graphs, these are drawn for only
        the non-parent edges that have already been represented (The edges that
        have not been represented will be handled in the recursive case).

        c. a child "L-style" connection on the lower right. Drawing of the
        children are handled recursively.

    4. The children of each node (wrt the directed DFS tree) are drawn
    underneath and to the right of it. In the case that a child node has already
    been drawn the connection is replaced with an ellipsis ("...") to indicate
    that there is one or more connections represented elsewhere.

    5. If a maximum depth is specified, an edge to nodes past this maximum
    depth will be represented by an ellipsis.

    Parameters
    ----------
    graph : nx.DiGraph | nx.Graph
        Graph to represent

    with_labels : bool | str
        If True will use the "label" attribute of a node to display if it
        exists otherwise it will use the node value itself. If given as a
        string, then that attribte name will be used instead of "label".
        Defaults to True.

    sources : List
        Specifies which nodes to start traversal from. Note: nodes that are not
        reachable from one of these sources may not be shown. If unspecified,
        the minimal set of nodes needed to reach all others will be used.

    max_depth : int | None
        The maximum depth to traverse before stopping. Defaults to None.

    ascii_only : Boolean
        If True only ASCII characters are used to construct the visualization

    Yields
    ------
    str : a line of generated text
    """
    is_directed = graph.is_directed()

    if is_directed:
        glyphs = AsciiDirectedGlyphs if ascii_only else UtfDirectedGlyphs
        succ = graph.succ
        pred = graph.pred
    else:
        glyphs = AsciiUndirectedGlyphs if ascii_only else UtfUndirectedGlyphs
        succ = graph.adj
        pred = graph.adj

    if isinstance(with_labels, str):
        label_attr = with_labels
    elif with_labels:
        label_attr = "label"
    else:
        label_attr = None

    if max_depth == 0:
        yield glyphs.empty + " ..."
    elif len(graph.nodes) == 0:
        yield glyphs.empty
    else:

        # If the nodes to traverse are unspecified, find the minimal set of
        # nodes that will reach the entire graph
        if sources is None:
            sources = _find_sources(graph)

        # Populate the stack with each:
        # 1. parent node in the DFS tree (or None for root nodes),
        # 2. the current node in the DFS tree
        # 2. a list of indentations indicating depth
        # 3. a flag indicating if the node is the final one to be written.
        # Reverse the stack so sources are popped in the correct order.
        last_idx = len(sources) - 1
        stack = [
            (None, node, [], (idx == last_idx)) for idx, node in enumerate(sources)
        ][::-1]

        num_skipped_children = defaultdict(lambda: 0)
        seen_nodes = set()
        while stack:
            parent, node, indents, this_islast = stack.pop()

            if node is not Ellipsis:
                skip = node in seen_nodes
                if skip:
                    # Mark that we skipped a parent's child
                    num_skipped_children[parent] += 1

                if this_islast:
                    # If we reached the last child of a parent, and we skipped
                    # any of that parents children, then we should emit an
                    # ellipsis at the end after this.
                    if num_skipped_children[parent] and parent is not None:

                        # Append the ellipsis to be emitted last
                        next_islast = True
                        try_frame = (node, Ellipsis, indents, next_islast)
                        stack.append(try_frame)

                        # Redo this frame, but not as a last object
                        next_islast = False
                        try_frame = (parent, node, indents, next_islast)
                        stack.append(try_frame)
                        continue

                if skip:
                    continue
                seen_nodes.add(node)

            if not indents:
                # Top level items (i.e. trees in the forest) get different
                # glyphs to indicate they are not actually connected
                if this_islast:
                    this_prefix = indents + [glyphs.newtree_last]
                    next_prefix = indents + [glyphs.endof_forest]
                else:
                    this_prefix = indents + [glyphs.newtree_mid]
                    next_prefix = indents + [glyphs.within_forest]

            else:
                # For individual tree edges distinguish between directed and
                # undirected cases
                if this_islast:
                    this_prefix = indents + [glyphs.last]
                    next_prefix = indents + [glyphs.endof_forest]
                else:
                    this_prefix = indents + [glyphs.mid]
                    next_prefix = indents + [glyphs.within_tree]

            if node is Ellipsis:
                label = " ..."
                suffix = ""
                children = []
            else:
                if label_attr is not None:
                    label = str(graph.nodes[node].get(label_attr, node))
                else:
                    label = str(node)

                # Determine:
                # (1) children to traverse into after showing this node.
                # (2) parents to immediately show to the right of this node.
                if is_directed:
                    # In the directed case we must show every successor node
                    # note: it may be skipped later, but we don't have that
                    # information here.
                    children = list(succ[node])
                    # In the directed case we must show every predecessor
                    # except for parent we directly traversed from.
                    handled_parents = {parent}
                else:
                    # Showing only the unseen children results in a more
                    # concise representation for the undirected case.
                    children = [
                        child for child in succ[node] if child not in seen_nodes
                    ]

                    # In the undirected case, parents are also children, so we
                    # only need to immediately show the ones we can no longer
                    # traverse
                    handled_parents = {*children, parent}

                if max_depth is not None and len(indents) == max_depth - 1:
                    # Use ellipsis to indicate we have reached maximum depth
                    if children:
                        children = [Ellipsis]
                    handled_parents = {parent}

                # The other parents are other predecessors of this node that
                # are not handled elsewhere.
                other_parents = [p for p in pred[node] if p not in handled_parents]
                if other_parents:
                    if label_attr is not None:
                        other_parents_labels = ", ".join(
                            [
                                str(graph.nodes[p].get(label_attr, p))
                                for p in other_parents
                            ]
                        )
                    else:
                        other_parents_labels = ", ".join(
                            [str(p) for p in other_parents]
                        )
                    suffix = " ".join(["", glyphs.backedge, other_parents_labels])
                else:
                    suffix = ""

            # Emit the line for this node, this will be called for each node
            # exactly once.
            yield "".join(this_prefix + [label, suffix])

            # Push children on the stack in reverse order so they are popped in
            # the original order.
            for idx, child in enumerate(children[::-1]):
                next_islast = idx == 0
                try_frame = (node, child, next_prefix, next_islast)
                stack.append(try_frame)


@open_file(1, "w")
def write_network_text(
    graph,
    path=None,
    with_labels=True,
    sources=None,
    max_depth=None,
    ascii_only=False,
    end="\n",
):
    """Creates a nice text representation of a graph

    This works via a depth-first traversal of the graph and writing a line for
    each unique node encountered. Non-tree edges are written to the right of
    each node, and connection to a non-tree edge is indicated with an ellipsis.
    This representation works best when the input graph is a forest, but any
    graph can be represented.

    Parameters
    ----------
    graph : nx.DiGraph | nx.Graph
        Graph to represent

    path : string or file or callable or None
       Filename or file handle for data output.
       if a function, then it will be called for each generated line.
       if None, this will default to "sys.stdout.write"

    with_labels : bool | str
        If True will use the "label" attribute of a node to display if it
        exists otherwise it will use the node value itself. If given as a
        string, then that attribte name will be used instead of "label".
        Defaults to True.

    sources : List
        Specifies which nodes to start traversal from. Note: nodes that are not
        reachable from one of these sources may not be shown. If unspecified,
        the minimal set of nodes needed to reach all others will be used.

    max_depth : int | None
        The maximum depth to traverse before stopping. Defaults to None.

    ascii_only : Boolean
        If True only ASCII characters are used to construct the visualization

    end : string
        The line ending characater

    Example
    -------
    >>> graph = nx.balanced_tree(r=2, h=2, create_using=nx.DiGraph)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├─╼ 1
        │   ├─╼ 3
        │   └─╼ 4
        └─╼ 2
            ├─╼ 5
            └─╼ 6

    >>> # A near tree with one non-tree edge
    >>> graph.add_edge(5, 1)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├─╼ 1 ╾ 5
        │   ├─╼ 3
        │   └─╼ 4
        └─╼ 2
            ├─╼ 5
            │   └─╼  ...
            └─╼ 6

    >>> graph = nx.cycle_graph(5)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├── 1
        │   └── 2
        │       └── 3
        │           └── 4 ─ 0
        └──  ...

    >>> graph = nx.generators.barbell_graph(4, 2)
    >>> nx.write_network_text(graph)
    ╙── 4
        ├── 5
        │   └── 6
        │       ├── 7
        │       │   ├── 8 ─ 6
        │       │   │   └── 9 ─ 6, 7
        │       │   └──  ...
        │       └──  ...
        └── 3
            ├── 0
            │   ├── 1 ─ 3
            │   │   └── 2 ─ 0, 3
            │   └──  ...
            └──  ...

    >>> graph = nx.complete_graph(5, create_using=nx.Graph)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├── 1
        │   ├── 2 ─ 0
        │   │   ├── 3 ─ 0, 1
        │   │   │   └── 4 ─ 0, 1, 2
        │   │   └──  ...
        │   └──  ...
        └──  ...

    >>> graph = nx.complete_graph(3, create_using=nx.DiGraph)
    >>> nx.write_network_text(graph)
    ╙── 0 ╾ 1, 2
        ├─╼ 1 ╾ 2
        │   ├─╼ 2 ╾ 0
        │   │   └─╼  ...
        │   └─╼  ...
        └─╼  ...
    """
    if path is None:
        # The path is unspecified, write to stdout
        _write = sys.stdout.write
    elif hasattr(path, "write"):
        # The path is already an open file
        _write = path.write
    elif callable(path):
        # The path is a custom callable
        _write = path
    else:
        raise TypeError(type(path))

    for line in generate_network_text(
        graph,
        with_labels=with_labels,
        sources=sources,
        max_depth=max_depth,
        ascii_only=ascii_only,
    ):
        _write(line + end)


def _find_sources(graph):
    """
    Determine a minimal set of nodes such that the entire graph is reachable
    """
    # For each connected part of the graph, choose at least
    # one node as a starting point, preferably without a parent
    if graph.is_directed():
        # Choose one node from each SCC with minimum in_degree
        sccs = list(nx.strongly_connected_components(graph))
        # condensing the SCCs forms a dag, the nodes in this graph with
        # 0 in-degree correspond to the SCCs from which the minimum set
        # of nodes from which all other nodes can be reached.
        scc_graph = nx.condensation(graph, sccs)
        supernode_to_nodes = {sn: [] for sn in scc_graph.nodes()}
        # Note: the order of mapping differs between pypy and cpython
        # so we have to loop over graph nodes for consistency
        mapping = scc_graph.graph["mapping"]
        for n in graph.nodes:
            sn = mapping[n]
            supernode_to_nodes[sn].append(n)
        sources = []
        for sn in scc_graph.nodes():
            if scc_graph.in_degree[sn] == 0:
                scc = supernode_to_nodes[sn]
                node = min(scc, key=lambda n: graph.in_degree[n])
                sources.append(node)
    else:
        # For undirected graph, the entire graph will be reachable as
        # long as we consider one node from every connected component
        sources = [
            min(cc, key=lambda n: graph.degree[n])
            for cc in nx.connected_components(graph)
        ]
        sources = sorted(sources, key=lambda n: graph.degree[n])
    return sources


def graph_str(graph, with_labels=True, sources=None, write=None, ascii_only=False):
    """Creates a nice utf8 representation of a forest

    This function has been superseded by
    :func:`nx.readwrite.text.generate_network_text`, which should be used
    instead.

    Parameters
    ----------
    graph : nx.DiGraph | nx.Graph
        Graph to represent (must be a tree, forest, or the empty graph)

    with_labels : bool
        If True will use the "label" attribute of a node to display if it
        exists otherwise it will use the node value itself. Defaults to True.

    sources : List
        Mainly relevant for undirected forests, specifies which nodes to list
        first. If unspecified the root nodes of each tree will be used for
        directed forests; for undirected forests this defaults to the nodes
        with the smallest degree.

    write : callable
        Function to use to write to, if None new lines are appended to
        a list and returned. If set to the `print` function, lines will
        be written to stdout as they are generated. If specified,
        this function will return None. Defaults to None.

    ascii_only : Boolean
        If True only ASCII characters are used to construct the visualization

    Returns
    -------
    str | None :
        utf8 representation of the tree / forest

    Example
    -------
    >>> graph = nx.balanced_tree(r=2, h=3, create_using=nx.DiGraph)
    >>> print(nx.graph_str(graph))
    ╙── 0
        ├─╼ 1
        │   ├─╼ 3
        │   │   ├─╼ 7
        │   │   └─╼ 8
        │   └─╼ 4
        │       ├─╼ 9
        │       └─╼ 10
        └─╼ 2
            ├─╼ 5
            │   ├─╼ 11
            │   └─╼ 12
            └─╼ 6
                ├─╼ 13
                └─╼ 14


    >>> graph = nx.balanced_tree(r=1, h=2, create_using=nx.Graph)
    >>> print(nx.graph_str(graph))
    ╙── 0
        └── 1
            └── 2

    >>> print(nx.graph_str(graph, ascii_only=True))
    +-- 0
        L-- 1
            L-- 2
    """
    printbuf = []
    if write is None:
        _write = printbuf.append
    else:
        _write = write

    write_network_text(
        graph,
        _write,
        with_labels=with_labels,
        sources=sources,
        ascii_only=ascii_only,
        end="",
    )

    if write is None:
        # Only return a string if the custom write function was not specified
        return "\n".join(printbuf)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
from xdev.cli.main import main


if __name__ == '__main__':
    """
    CommandLine:
        xdev --help
        xdev --version
        xdev info
        xdev sed "main" "MAIN" "." --dry=True --include="*_*.py"
        xdev find "*_*.py"  '.'
        xdev codeblock "
            import sys
            print(sys.argv)
            print([
                'hello world'
            ])
        "
    """
    main()



import xinspect

get_func_kwargs = xinspect.get_func_kwargs


def get_stack_frame(N=0, strict=True):
    """
    Args:
        N (int): N=0 means the frame you called this function in.
                 N=1 is the parent frame.
        strict (bool): (default = True)
    """
    import inspect
    frame_cur = inspect.currentframe()
    for idx in range(N + 1):
        # always skip the frame of this function
        frame_next = frame_cur.f_back
        if frame_next is None:
            if strict:
                raise AssertionError('Frame level {:!r} is root'.format(idx))
            else:
                break
        frame_cur = frame_next
    return frame_cur


def distext(obj):
    """
    Like dis.dis, but returns the text

    Args:
        obj : a compiled object (e.g. a function, class, generator, etc...)

    Returns:
        str: text of the python byte code

    Example:
        >>> from xdev.introspect import *  # NOQA
        >>> import ubelt as ub
        >>> code = ub.codeblock(
            '''
            def foo(a, b):
                print('hello')
                if __debug__:
                    if a is None or b is None:
                        raise ValueError
                print('world')
                # This is not entirely optimized away
                if __debug__ and (a is None or b is None):
                    raise ValueError
                return a + b
            ''')
        >>> obj = compile(code, filename='<memory>', mode='exec', dont_inherit=True, optimize=0)
        >>> print(' ** UNOPTIMIZED')
        >>> text = distext(obj)
        >>> print(text)
        >>> print(' ** OPTIMIZED')
        >>> obj = compile(code, filename='<memory>', mode='exec', dont_inherit=True, optimize=1)
        >>> text = distext(obj)
        >>> print(text)
    """
    import dis
    import io
    file = io.StringIO()
    dis.dis(obj, file=file)
    file.seek(0)
    text = file.read()
    return text


# def import_star(module):
#     """
#     This is the unholy grail

#     Import all contents of the module into the global scope

#     Args:
#         module (str | module): a path, module name, or module itself

#     Example:
#         >>> # xdoctest: +DISABLE_DOCTEST
#         >>> from xdev.misc import import_star
#         >>> module = 'ubelt'
#         >>> print(import_star(module).keys())
#     """
#     import ubelt as ub
#     if isinstance(module, str):
#         if exists(module):
#             module = ub.import_module_from_path(module)
#         else:
#             module = ub.import_module_from_name(module)

#     mod_attrs = {
#         key: val for key, val in module.__dict__.items()
#         if not key.startswith('_')
#     }
#     parent_frame = get_stack_frame(N=2)
#     print('parent_frame.f_code.co_names = {!r}'.format(parent_frame.f_code.co_names))
#     print('parent_frame.f_code.co_name = {!r}'.format(parent_frame.f_code.co_name))
#     # Note: locals is usually read only
#     # parent_frame.f_locals.update(mod_attrs)
#     parent_frame.f_globals.update(mod_attrs)
#     return mod_attrs


def iter_object_tree(obj):
    # ub.IndexableWalker
    data = obj.__dict__
    prefix = []

    list_cls = (list, tuple)
    dict_cls = (dict,)
    indexable_cls = dict_cls + list_cls

    seen_ = set()
    stack = [(data, prefix)]
    while stack:
        _data, _prefix = stack.pop()
        # Create an items iterable of depending on the indexable data type
        if hasattr(_data, '__dict__'):
            items = _data.__dict__.items()
        elif isinstance(_data, list_cls):
            items = enumerate(_data)
        elif isinstance(_data, dict_cls):
            items = _data.items()
        else:
            raise TypeError(type(_data))

        for key, value in items:
            # Yield the full path to this position and its value
            path = _prefix + [key]

            message = yield path, value
            # If the value at this path is also indexable, then continue
            # the traversal, unless the False message was explicitly sent
            # by the caller.
            if message is False:
                # Because the `send` method will return the next value,
                # we yield a dummy value so we don't clobber the next
                # item in the traversal.
                yield None
            else:
                if isinstance(value, indexable_cls) or hasattr(value, '__dict__'):
                    objid = id(value)
                    if objid not in seen_:
                        seen_.add(objid)
                        stack.append((value, path))


def test_object_pickleability(obj):
    # ub.IndexableWalker

    import pickle
    serialized = pickle.dumps(obj)
    recon = pickle.loads(serialized)

    data = obj.__dict__
    prefix = []

    list_cls = (list, tuple)
    dict_cls = (dict,)
    indexable_cls = dict_cls + list_cls

    def condition(path, value):
        import pickle
        dumped = pickle.dumps(value)
        try:
            recon = pickle.loads(dumped)
        except Exception:
            return True
        else:
            return False
        ignore_types = (float, int, bool, type(None))
        return not isinstance(value, ignore_types)

    walker = iter_object_tree(obj)
    for path, value in walker:
        flag = condition(path, value)
        if flag:
            print('BAD FLAG path = {!r}'.format(path))
            walker.send(False)

    found = []


    keyfn = type

    seen_ = set()

    stack = [(data, prefix)]
    while stack:
        _data, _prefix = stack.pop()
        # Create an items iterable of depending on the indexable data type
        if isinstance(_data, list_cls):
            items = enumerate(_data)
        elif isinstance(_data, dict_cls):
            items = _data.items()
        elif hasattr(_data, '__dict__'):
            items = _data.__dict__.items()
        else:
            raise TypeError(type(_data))

        for key, value in items:
            # Yield the full path to this position and its value
            path = _prefix + [key]
            flag = condition(path, value)
            if flag:
                store = keyfn(value)
                found.append((path, store))

                if isinstance(value, indexable_cls) or hasattr(value, '__dict__'):
                    objid = id(value)
                    if objid not in seen_:
                        seen_.add(objid)
                        stack.append((value, path))



def bubbletext(text, font='cybermedium'):
    """
    Uses pyfiglet to create bubble text.

    Args:
        font (str): default=cybermedium, other fonts include: cybersmall and
            cyberlarge.

    References:
        http://www.figlet.org/

    Example:
        >>> bubble_text = bubbletext('TESTING BUBBLE TEXT', font='cybermedium')
        >>> print(bubble_text)
    """
    try:
        import pyfiglet
    except ImportError:
        print('Missing pyfiglet')
        return text
    else:
        bubble_text = pyfiglet.figlet_format(text, font=font)
        return bubble_text


def conj_phrase(list_, cond='or'):
    """
    Joins a list of words using English conjunction rules

    Args:
        list_ (list):  of strings
        cond (str): a conjunction (or, and, but)

    Returns:
        str: the joined cconjunction phrase

    References:
        http://en.wikipedia.org/wiki/Conjunction_(grammar)

    Example:
        >>> list_ = ['a', 'b', 'c']
        >>> result = conj_phrase(list_, 'or')
        >>> print(result)
        a, b, or c

    Example1:
        >>> list_ = ['a', 'b']
        >>> result = conj_phrase(list_, 'and')
        >>> print(result)
        a and b
    """
    if len(list_) == 0:
        return ''
    elif len(list_) == 1:
        return list_[0]
    elif len(list_) == 2:
        return ' '.join((list_[0], cond, list_[1]))
    else:
        condstr = ''.join((', ' + cond, ' '))
        return ', '.join((', '.join(list_[:-2]), condstr.join(list_[-2:])))


def take_column(list_, colx):
    """ iterator version of take_column """
    if isinstance(colx, list):
        # multi select
        return [[row[colx_] for colx_ in colx] for row in list_]
    else:
        return [row[colx] for row in list_]



"""
Functions related to interacting with data via an OS Desktop GUI.
"""
from os.path import normpath
from os.path import exists
from os.path import sys
from six import types
import os
import six
import ubelt as ub


def editfile(fpath, verbose=True):
    """
    Opens a file or code corresponding to a live python object in your
    preferred visual editor. This function is mainly useful in an interactive
    IPython session.

    The visual editor is determined by the `VISUAL` environment variable.  If
    this is not specified it defaults to gvim.

    Args:
        fpath (PathLike): a file path or python module / function
        verbose (int): verbosity

    DisableExample:
        >>> # This test interacts with a GUI frontend, not sure how to test.
        >>> import xdev
        >>> ub.editfile(xdev.misc.__file__)
        >>> ub.editfile(xdev)
        >>> ub.editfile(xdev.editfile)
    """
    if not isinstance(fpath, six.string_types):
        if isinstance(fpath, types.ModuleType):
            fpath = fpath.__file__
        else:
            fpath =  sys.modules[fpath.__module__].__file__
        fpath_py = fpath.replace('.pyc', '.py')
        if exists(fpath_py):
            fpath = fpath_py

    if verbose:
        print('[xdev] editfile("{}")'.format(fpath))

    editor = os.environ.get('VISUAL', 'gvim')
    if not ub.find_exe(editor):
        import warnings
        warnings.warn('Cannot find visual editor={}'.format(editor), UserWarning)
        # Try and fallback on commonly installed editor
        alt_candidates = [
            'gedit',
            'TextEdit'
            'Notepad',
        ]
        for cand in alt_candidates:
            if ub.find_exe(cand):
                editor = cand

    if not exists(fpath):
        raise IOError('Cannot start nonexistant file: %r' % fpath)
    ub.cmd([editor, fpath], fpath, detach=True)


def view_directory(dpath=None, verbose=False):
    """
    View a directory in the operating system file browser. Currently supports
    windows explorer, mac open, and linux nautlius.

    Args:
        dpath (PathLike | None): directory name
        verbose (bool): verbosity
    """
    if dpath is None:
        dpath = os.getcwd()
    dpath = os.path.normpath(dpath)
    if verbose:
        print('[xdev] view_directory({!r}) '.format(dpath))
    if not exists(dpath):
        raise Exception('Cannot view nonexistant directory: {!r}'.format(dpath))
    if False:
        try:
            import vimtk.xctrl
            import vimtk.cplat_ctrl
            if vimtk.xctrl.is_directory_open(dpath):
                if verbose:
                    print('[xdev] dpath={!r} is already open'.format(dpath))
                win = vimtk.cplat_ctrl.Window.find('Nautilus.*' + os.path.basename(dpath))
                win.focus()
                return
        except Exception:
            pass
    if ub.LINUX:
        info = ub.cmd(('nautilus', dpath), detach=True, verbose=verbose)
    elif ub.DARWIN:
        info = ub.cmd(('open', dpath), detach=True, verbose=verbose)
    elif ub.WIN32:
        info = ub.cmd(('explorer.exe', dpath), detach=True, verbose=verbose)
    else:
        raise RuntimeError('Unknown Platform')
    if info is not None:
        if not info['proc']:
            raise Exception('startfile failed')


def startfile(fpath, verbose=True):
    """
    Uses default program defined by the system to open a file.
    This is done via `os.startfile` on windows, `open` on mac, and `xdg-open`
    on linux.

    Args:
        fpath (PathLike): a file to open using the program associated with the
            files extension type.
        verbose (int): verbosity

    References:
        http://stackoverflow.com/questions/2692873/quote-posix

    DisableExample:
        >>> # This test interacts with a GUI frontend, not sure how to test.
        >>> import ubelt as ub
        >>> base = ub.ensure_app_cache_dir('ubelt')
        >>> fpath1 = join(base, 'test_open.txt')
        >>> ub.touch(fpath1)
        >>> proc = ub.startfile(fpath1)
    """
    if verbose:
        print('[xdev] startfile("{}")'.format(fpath))
    fpath = normpath(fpath)
    if not exists(fpath):
        raise Exception('Cannot start nonexistant file: {!r}'.format(fpath))
    if not ub.WIN32:
        import shlex
        fpath = shlex.quote(fpath)
    if ub.LINUX:
        info = ub.cmd(('xdg-open', fpath), detach=True, verbose=verbose)
    elif ub.DARWIN:
        info = ub.cmd(('open', fpath), detach=True, verbose=verbose)
    elif ub.WIN32:
        os.startfile(fpath)
        info = None
    else:
        raise RuntimeError('Unknown Platform')
    if info is not None:
        if not info['proc']:
            raise Exception('startfile failed')



"""
This is Jon Crall's xdev module.

These are tools I often use in IPython, but they almost never make it into
production code, otherwise they would be in :mod:`ubelt`.
"""

__dev__ = """
CommandLine:
    # Regenerate the tail of this file
    mkinit ~/code/xdev/xdev -w --lazy

TODO:
    - [ ] Update mkinit so we can either:
        (1) blocklist specific modules from importing their attrs or
        (2) passlist the modules that will import their attrs

    - [ ] Perhaps let submodules specify a 2-tuple with the second item
        being a dict that indicates: (nomod, noattr)?

    - [ ] Automatically add custom defined names in this file to __all__
"""

__version__ = '1.1.0'


# __submodules__ = [
#     'embeding',
#     'interactive_iter',
#     'desktop_interaction',
#     'introspect',
#     'class_reloader',
#     'search_replace',
#     'misc',
#     'util_path',
#     'patterns',
#     'autojit',
#     'profiler',
#     'tracebacks',
# ]

# __extra_all__ = [
#     'util'
# ]


from xdev import util

### The following is autogenerated


def lazy_import(module_name, submodules, submod_attrs):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    if os.environ.get('EAGER_IMPORT', ''):
        for name in submodules:
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'algo',
        'autojit',
        'class_reloader',
        'cli',
        'desktop_interaction',
        'embeding',
        'format_quotes',
        'interactive_iter',
        'introspect',
        'misc',
        'patterns',
        'profiler',
        'regex_builder',
        'search_replace',
        'tracebacks',
        'util',
        'util_networkx',
        'util_path',
    },
    submod_attrs={
        'algo': [
            'edit_distance',
            'knapsack',
            'knapsack_greedy',
            'knapsack_ilp',
            'knapsack_iterative',
            'knapsack_iterative_int',
            'knapsack_iterative_numpy',
            'number_of_decimals',
        ],
        'autojit': [
            'import_module_from_pyx',
        ],
        'class_reloader': [
            'reload_class',
        ],
        'desktop_interaction': [
            'editfile',
            'startfile',
            'view_directory',
        ],
        'embeding': [
            'EmbedOnException',
            'embed',
            'embed_on_exception_context',
            'fix_embed_globals',
        ],
        'format_quotes': [
            'format_quotes',
            'format_quotes_in_file',
            'format_quotes_in_text',
        ],
        'interactive_iter': [
            'InteractiveIter',
        ],
        'introspect': [
            'distext',
            'get_func_kwargs',
            'get_stack_frame',
            'iter_object_tree',
            'test_object_pickleability',
        ],
        'misc': [
            'byte_str',
            'difftext',
            'nested_type',
            'quantum_random',
            'set_overlaps',
            'tree_repr',
        ],
        'patterns': [
            'MultiPattern',
            'Pattern',
            'PatternBase',
            'RE_Pattern',
            'our_extended_regex_compile',
        ],
        'profiler': [
            'IS_PROFILING',
            'profile',
            'profile_globals',
            'profile_now',
        ],
        'regex_builder': [
            'PythonRegexBuilder',
            'RegexBuilder',
            'VimRegexBuilder',
        ],
        'search_replace': [
            'GrepResult',
            'find',
            'grep',
            'grepfile',
            'greptext',
            'sed',
            'sedfile',
        ],
        'tracebacks': [
            'make_warnings_print_tracebacks',
        ],
        'util': [
            'bubbletext',
            'conj_phrase',
            'take_column',
        ],
        'util_networkx': [
            'AsciiDirectedGlyphs',
            'AsciiUndirectedGlyphs',
            'UtfDirectedGlyphs',
            'UtfUndirectedGlyphs',
            'generate_network_text',
            'graph_str',
            'write_network_text',
        ],
        'util_path': [
            'ChDir',
            'sidecar_glob',
            'tree',
        ],
    },
)


def __dir__():
    return __all__

__all__ = ['AsciiDirectedGlyphs', 'AsciiUndirectedGlyphs', 'ChDir',
           'EmbedOnException', 'GrepResult', 'IS_PROFILING', 'InteractiveIter',
           'MultiPattern', 'Pattern', 'PatternBase', 'PythonRegexBuilder',
           'RE_Pattern', 'RegexBuilder', 'UtfDirectedGlyphs',
           'UtfUndirectedGlyphs', 'VimRegexBuilder', 'algo', 'autojit',
           'bubbletext', 'byte_str', 'class_reloader', 'cli', 'conj_phrase',
           'desktop_interaction', 'difftext', 'distext', 'edit_distance',
           'editfile', 'embed', 'embed_on_exception_context', 'embeding',
           'find', 'fix_embed_globals', 'format_quotes',
           'format_quotes_in_file', 'format_quotes_in_text',
           'generate_network_text', 'get_func_kwargs', 'get_stack_frame',
           'graph_str', 'grep', 'grepfile', 'greptext',
           'import_module_from_pyx', 'interactive_iter', 'introspect',
           'iter_object_tree', 'knapsack', 'knapsack_greedy', 'knapsack_ilp',
           'knapsack_iterative', 'knapsack_iterative_int',
           'knapsack_iterative_numpy', 'make_warnings_print_tracebacks',
           'misc', 'nested_type', 'number_of_decimals',
           'our_extended_regex_compile', 'patterns', 'profile',
           'profile_globals', 'profile_now', 'profiler', 'quantum_random',
           'regex_builder', 'reload_class', 'search_replace', 'sed', 'sedfile',
           'set_overlaps', 'sidecar_glob', 'startfile', 'take_column',
           'test_object_pickleability', 'tracebacks', 'tree', 'tree_repr',
           'util', 'util_networkx', 'util_path', 'view_directory',
           'write_network_text']



import sys
from IPython.terminal.embed import InteractiveShellEmbed, load_default_config
# from IPython.terminal.ipapp import load_default_config


class InteractiveShellEmbedEnhanced(InteractiveShellEmbed):
    """
    A workaround to issue https://github.com/ipython/ipython/issues/10695
    In embedded ipython shell, new variables are created in locals() instead of globals()
    and functions do not make closures for the locals.

    Examples:

    If not embedded from a function, then nothing special needs to be done other than calling embed() defined below.
    These code will work:

        a=3; (lambda: a)()
        import time; (lambda: time.time())()

    If embedded from a function, in order to introduce new variables to global name space, we need to call
    ipy.to_global() first and then call ipy.to_local() to come back

        ipy.to_global()
        a=3  # here we cannot access the local variables
        ipy.to_local()
        (lambda: a)()

    It seems that a contextmanager to switch back automatically will not work:

        with ipy.to_global_manager():
            a=3
            print((lambda: a)())

    It may be because of that IPython compiles the whole cell and run it using exec(code, globals, locals).
    But globals and locals are calculated at the time this context manager is defined.
    """
    @property
    def user_global_ns(self):
        if getattr(self, 'embedded_outside_func', False):
            return self.user_ns
        else:
            return self.user_module.__dict__

    def init_frame(self, frame):
        if frame.f_code.co_name == '<module>':  # if embedded outside a function
            self.embedded_outside_func = True
        else:  # if embedded from a function
            self.embedded_outside_func = False

            self._saved_user_ns = None
            self._ns = {}
            # make global variables for user access to the histories
            self._ns['_ih'] = self.history_manager.input_hist_parsed
            self._ns['_oh'] = self.history_manager.output_hist
            self._ns['_dh'] = self.history_manager.dir_hist

            # user aliases to input and output histories.  These shouldn't show up
            # in %who, as they can have very large reprs.
            self._ns['In']  = self.history_manager.input_hist_parsed
            self._ns['Out'] = self.history_manager.output_hist

            # Store myself as the public api!!!
            # ns['get_ipython'] = self.get_ipython
            self._ns['exit'] = self.exiter
            self._ns['quit'] = self.exiter

            self.to_global = self._to_global
            self.to_local = self._to_local

    def _to_global(self):
        self._saved_user_ns = self.user_ns
        assert not set(self._ns).intersection(self.user_module.__dict__)
        self.user_ns = self.user_module.__dict__
        self.user_ns.update(self._ns)

    def _to_local(self):
        if self._saved_user_ns is None:
            return
        for key in self._ns:
            del self.user_ns[key]
        self.user_ns = self._saved_user_ns

    def share_locals(self):
        """
        share the locals to global manually
        """
        if (id(self.user_ns) != id(self.user_module.__dict__)):
            self.user_module.__dict__.update(self.user_ns)


def embed2(local_ns=None, **kwargs):
    """
    based on IPython.terminal.embed.embed()

    References:
        https://github.com/ipython/ipython/issues/10695
    """
    import glob
    import os
    from IPython.core.interactiveshell import InteractiveShell
    config = kwargs.get('config')
    header = kwargs.pop('header', u'')
    compile_flags = kwargs.pop('compile_flags', None)
    if config is None:
        config = load_default_config()
        config.InteractiveShellEmbedEnhanced = config.TerminalInteractiveShell
        kwargs['config'] = config
    #save ps1/ps2 if defined
    ps1 = None
    ps2 = None
    try:
        ps1 = sys.ps1
        ps2 = sys.ps2
    except AttributeError:
        pass
    #save previous instance
    saved_shell_instance = InteractiveShell._instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
    frame = sys._getframe(1)

    # shell is the ipython instance returned from get_ipython()
    # frame refers to the caller of this function
    shell = InteractiveShellEmbedEnhanced.instance(_init_location_id='%s:%s' % (
        frame.f_code.co_filename, frame.f_lineno), **kwargs)
    shell.init_frame(frame)

    #######################
    # load the startup files and update the local_ns
    ######################

    # if local_ns is None:
    #     local_ns = frame.f_locals
    global_ns = frame.f_globals
    global_ns['ipy'] = shell
    global_ns['share_locals'] = shell.share_locals

    startup_files = glob.glob(os.path.join(shell.profile_dir.startup_dir, '*.py'))
    startup_files += glob.glob(os.path.join(shell.profile_dir.startup_dir, '*.ipy'))

    if '__file__' in global_ns:
        hasfile = True
        cfile = global_ns['__file__']
    else:
        hasfile = False

    for filename in sorted(startup_files):
        if filename.endswith('.ipy'):
            shell.safe_execfile_ipy(filename)
        else:
            global_ns['__file__'] = filename
            shell.safe_execfile(filename, global_ns, raise_exceptions=True)  # this updates the global_ns

    if hasfile:
        global_ns['__file__'] = cfile
    else:
        del global_ns['__file__']

    ########################
    #  launch the shell
    #######################
    shell(local_ns=local_ns, header=header, stack_depth=2, compile_flags=compile_flags,
          _call_location_id='%s:%s' % (frame.f_code.co_filename, frame.f_lineno))
    InteractiveShellEmbedEnhanced.clear_instance()
    #restore previous instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
        for subclass in cls._walk_mro():
            subclass._instance = saved_shell_instance
    if ps1 is not None:
        sys.ps1 = ps1
        sys.ps2 = ps2



"""
An encapsulation of regex and glob (and maybe other) patterns.
"""
import os
import re
import fnmatch
import ubelt as ub
import pathlib
from . import util_path

if hasattr(re, 'Pattern'):
    RE_Pattern = re.Pattern
else:
    # sys.version_info[0:2] <= 3.6
    RE_Pattern = type(re.compile('.*'))


class PatternBase:

    def match(self, text):
        raise NotImplementedError

    def search(self, text):
        raise NotImplementedError

    def sub(self, repl, text):
        raise NotImplementedError


# # TODO: wrapper for results of re and fnmatch
# class Match(ub.NiceRepr):
#     def __init__(self, string, pos, endpos):
#         self.string = string
#         self.pos = pos
#         self.endpos = endpos
#         self.lastgroup
#         self.lastindex
#     def span(self):
#         return (self.pos, self.endpos)
#     def __nice__(self):
#         return self.string
#     def __bool__(self):
#         return True


def our_extended_regex_compile():
    r"""
    Adds suppport for vim-like \< and \> syntax to represent \b
    """
    pass


class Pattern(PatternBase, ub.NiceRepr):
    """
    Provides a common API to several common pattern matching syntaxes.

    A general patterns class, which can be strict, regex, or glob.

    Args:
        pattern (str | object):
            The pattern text or a precompiled backend pattern object

        backend (str):
            Code indicating what backend the pattern text should be
            interpereted with. Current modes are: strict, regex, and glob.

    Notes:
        The glob backend uses the :module:`fnmatch` module [fnmatch_docs]_.
        The regex backend uses the Python :module:`re` module.
        The strict backend uses the "==" string equality testing.

    References:
        ..[fnmatch_docs] https://docs.python.org/3/library/fnmatch.html

    Example:
        >>> repat = Pattern.coerce('foo.*', 'regex')
        >>> assert repat.match('foobar')
        >>> assert not repat.match('barfoo')
        >>> globpat = Pattern.coerce('foo*', 'glob')
        >>> assert globpat.match('foobar')
        >>> assert not globpat.match('barfoo')
        >>> globpat = Pattern.coerce('[foo|bar]', 'glob')
        >>> globpat.match('foo')
        >>> repat = Pattern.coerce('foo.*', 'regex')
        >>> match = repat.search('baz-biz-foobar')
        >>> match = repat.match('baz-biz-foobar')
    """
    def __init__(self, pattern, backend):
        if isinstance(pattern, pathlib.Path):
            pattern = os.fspath(pattern)
        if backend == 'regex':
            if isinstance(pattern, str):
                pattern = re.compile(pattern)
        self.pattern = pattern
        self.backend = backend

    def __nice__(self):
        return '{}, {}'.format(self.pattern, self.backend)

    def to_regex(self):
        """
        Returns an equivalent pattern with the regular expression backend

        Example:
            >>> globpat = Pattern.coerce('foo*', 'glob')
            >>> strictpat = Pattern.coerce('foo*', 'strict')
            >>> repat1 = strictpat.to_regex()
            >>> repat2 = globpat.to_regex()
            >>> print(f'repat1={repat1}')
            >>> print(f'repat2={repat2}')
        """
        if self.backend == 'regex':
            regex_pattern = self.pattern
        elif self.backend == 'glob':
            regex_pattern = fnmatch.translate(self.pattern)
        elif self.backend == 'strict':
            regex_pattern = re.escape(self.pattern)
        else:
            raise AssertionError
        new = self.__class__(regex_pattern, 'regex')
        return new

    @classmethod
    def from_regex(cls, data, flags=0, multiline=False, dotall=False,
                   ignorecase=False):
        if multiline:
            flags |= re.MULTILINE
        if dotall:
            flags |= re.DOTALL
        if multiline:
            flags |= re.DOTALL
        if ignorecase:
            flags |= re.IGNORECASE
        pat = re.compile(data, flags=flags)
        self = cls(pat, 'regex')
        return self

    @classmethod
    def from_glob(cls, data):
        self = cls(data, 'glob')
        return self

    @classmethod
    def coerce_backend(cls, data, hint='glob'):
        if isinstance(data, RE_Pattern):
            backend = 'regex'
        elif isinstance(data, cls) or type(data).__name__ == cls.__name__:
            backend = data.backend
        else:
            backend = hint
        return backend

    def match(self, text):
        if self.backend == 'regex':
            return self.pattern.match(text)
        elif self.backend == 'glob':
            return fnmatch.fnmatch(text, self.pattern)
        elif self.backend == 'strict':
            return self.pattern == text
        else:
            raise KeyError(self.backend)

    def search(self, text):
        if self.backend == 'regex':
            return self.pattern.search(text)
        elif self.backend == 'glob':
            return fnmatch.fnmatch(text, '*{}*'.format(self.pattern))
        elif self.backend == 'strict':
            return self.pattern in text
        else:
            raise KeyError(self.backend)

    def sub(self, repl, text, count=-1):
        """
        Args:
            repl (str): text to insert in place of pattern
            text (str): text to be searched and modified
            count (int): if non-negative, the maximum number of replacements
                that will be made.
        """
        if count == 0:
            return text  # make regex conform to the API
        if self.backend == 'regex':
            return self.pattern.sub(repl, text, count=max(0, count))
        elif self.backend == 'glob':
            raise NotImplementedError
        elif self.backend == 'strict':
            return text.replace(self.pattern, repl, count=count)
        else:
            raise KeyError(self.backend)

    @classmethod
    def coerce(cls, data, hint='glob'):
        """
        Example:
            >>> pat = Pattern.coerce('foo*', 'glob')
            >>> pat2 = Pattern.coerce(pat, 'regex')
            >>> print('pat = {}'.format(ub.repr2(pat, nl=1)))
            >>> print('pat2 = {}'.format(ub.repr2(pat2, nl=1)))

            Pattern.coerce(['a', 'b', 'c'])
        """
        if isinstance(data, cls) or type(data).__name__ == cls.__name__:
            self = data
        else:
            # string
            backend = cls.coerce_backend(data, hint=hint)
            self = cls(data, backend)
        return self

    def paths(self, cwd=None, recursive=False):
        """
        Find paths in the filesystem that match this pattern

        Yields:
            ub.Path
        """
        if self.backend == 'glob':
            import glob
            with util_path.ChDir(cwd):
                yield from map(ub.Path, glob.glob(
                    self.pattern, recursive=recursive))
        else:
            raise NotImplementedError


class MultiPattern(PatternBase, ub.NiceRepr):
    """
    Example:
        >>> dpath = ub.Path.appdir('xdev/tests/multipattern_paths').ensuredir().delete().ensuredir()
        >>> (dpath / 'file0.txt').touch()
        >>> (dpath / 'data0.dat').touch()
        >>> (dpath / 'other0.txt').touch()
        >>> ((dpath / 'dir1').ensuredir() / 'file1.txt').touch()
        >>> ((dpath / 'dir2').ensuredir() / 'file2.txt').touch()
        >>> ((dpath / 'dir2').ensuredir() / 'file3.txt').touch()
        >>> ((dpath / 'dir1').ensuredir() / 'data.dat').touch()
        >>> ((dpath / 'dir2').ensuredir() / 'data.dat').touch()
        >>> ((dpath / 'dir2').ensuredir() / 'data.dat').touch()
        >>> pat = MultiPattern.coerce(['*.txt'], 'glob')
        >>> print(list(pat.paths(cwd=dpath)))
        >>> pat = MultiPattern.coerce(['*0*', '**/*.txt'], 'glob')
        >>> print(list(pat.paths(cwd=dpath, recursive=1)))
        >>> pat = MultiPattern.coerce(['*.txt', '**/*.txt', '**/*.dat'], 'glob')
        >>> print(list(pat.paths(cwd=dpath)))
    """
    def __init__(self, patterns, predicate):
        self.predicate = predicate
        self.patterns = patterns

    def __nice__(self):
        return f'{self.predicate.__name__}({[str(p) for p in self.patterns]})'

    def match(self, text):
        return self.predicate(p.match(text) for p in self.patterns)

    def paths(self, cwd=None, recursive=False):
        groups = (p.paths(cwd=cwd, recursive=recursive) for p in self.patterns)
        if self.predicate in {any}:  # all}:
            yield from ub.unique(ub.flatten(groups))
        elif self.predicate in {all}:  # all}:
            yield from set.intersection(*map(set, groups))
        else:
            raise NotImplementedError

    # def search(self, text):
    #     return self.predicate(p.search(text) for p in self.patterns)

    def _squeeze(self):
        if self.predicate in {any, all}:
            if len(self.patterns) == 1:
                new = self.patterns[0]
            else:
                new = self
        else:
            raise NotImplementedError
        return new

    @classmethod
    def coerce(cls, data, hint='glob', predicate='any'):
        """
        Example:
            >>> pat = MultiPattern.coerce('foo*', 'glob')
            >>> pat2 = MultiPattern.coerce(pat, 'regex')
            >>> pat3 = MultiPattern.coerce([pat, pat], 'regex')
            >>> print('pat = {}'.format(ub.repr2(pat, nl=1)))
            >>> print('pat2 = {}'.format(ub.repr2(pat2, nl=1)))
            >>> print('pat3 = {!r}'.format(pat3))

            Pattern.coerce(['a', 'b', 'c'])
        """
        if isinstance(data, cls) or type(data).__name__ == cls.__name__:
            self = data
        else:
            # coerce predicate
            if predicate == 'any':
                predicate = any
            else:
                raise NotImplementedError
            if isinstance(data, str):
                backend = Pattern.coerce_backend(data, hint=hint)
                pat = Pattern.coerce(data, backend)
                patterns = [pat]
                self = MultiPattern(patterns, predicate)
            else:
                self = MultiPattern([
                    MultiPattern.coerce(d, hint)._squeeze() for d in data], predicate)
        return self



"""
Utilities to just-in-time-cythonize a module at runtime.
"""
from collections import defaultdict
from os.path import dirname, join, basename, splitext, exists
import os
import warnings


__all__ = [
    'import_module_from_pyx',
]

# Track the number of times we've tried to autojit specific pyx files
NUM_AUTOJIT_TRIES = defaultdict(lambda: 0)
# MAX_AUTOJIT_TRIES = 1
MAX_AUTOJIT_TRIES = float('inf')
# 1


def import_module_from_pyx(fname, dpath=None, error="raise", autojit=True,
                           verbose=1, recompile=False, annotate=False):
    """
    Attempts to import a module corresponding to a pyx file.

    If the corresponding compiled module is not found, this can attempt to
    JIT-cythonize the pyx file.

    Parameters
    ----------
    fname : str
        The basename of the cython pyx file

    dpath : str
        The directory containing the cython pyx file

    error : str
        Can be "raise" or "ignore"

    autojit : bool
        If True, we will cythonize and compile the pyx file if possible.

    verbose : int
        verbosity level (higher is more verbose)

    recompile : bool
        if True force recompile

    Returns
    -------
    ModuleType | None : module
        Returns the compiled and imported module if possible, otherwise None
    """
    if dpath is None:
        dpath = dirname(fname)

    pyx_fpath = join(dpath, fname)

    if verbose > 3:
        print('[xdev.import] pyx_fpath = {!r}'.format(pyx_fpath))

    if not exists(pyx_fpath):
        raise AssertionError("pyx file {!r} does not exist".format(pyx_fpath))

    try:
        # This functionality depends on ubelt
        # TODO: the required functionality could be moved to nx.utils
        import ubelt as ub
    except Exception:
        if verbose:
            print("Autojit requires ubelt, which failed to import")
        if error == "ignore":
            module = None
        elif error == "raise":
            raise
        else:
            raise KeyError(error)
    else:

        if autojit:
            if verbose > 3:
                print('autojit = {!r}'.format(autojit))
            # Try to JIT the cython module if we ship the pyx without the compiled
            # library.
            NUM_AUTOJIT_TRIES[pyx_fpath] += 1
            if verbose > 3:
                print('NUM_AUTOJIT_TRIES = {!r}'.format(NUM_AUTOJIT_TRIES))
                print('MAX_AUTOJIT_TRIES = {!r}'.format(MAX_AUTOJIT_TRIES))
            if recompile or NUM_AUTOJIT_TRIES[pyx_fpath] <= MAX_AUTOJIT_TRIES:
                if verbose > 3:
                    print('try it')
                try:
                    _autojit_cython(pyx_fpath, verbose=verbose,
                                    recompile=recompile, annotate=annotate)
                    if verbose > 3:
                        print('got it')
                except Exception as ex:
                    warnings.warn("Cython autojit failed: ex={!r}".format(ex))
                    if error == "raise":
                        raise

        try:
            # if recompile:
            #     # Doesnt work. The module is not reloaded.
            #     modname = ub.modpath_to_modname(pyx_fpath)
            #     import sys
            #     sys.modules.pop(modname, None)
            module = ub.import_module_from_path(pyx_fpath)
        except Exception:
            if error == "ignore":
                module = None
            elif error == "raise":
                raise
            else:
                raise KeyError(error)

        return module


def _platform_pylib_exts():  # nocover
    """
    Returns .so, .pyd, or .dylib depending on linux, win or mac.  Returns the
    previous with and without abi (e.g. .cpython-35m-x86_64-linux-gnu) flags.
    """
    import sysconfig

    valid_exts = []
    # handle PEP 3149 -- ABI version tagged .so files
    base_ext = "." + sysconfig.get_config_var("EXT_SUFFIX").split(".")[-1]
    # ABI = application binary interface
    tags = [
        sysconfig.get_config_var("SOABI"),
        "abi3",  # not sure why this one is valid, but it is
    ]
    tags = [t for t in tags if t]
    for tag in tags:
        valid_exts.append("." + tag + base_ext)
    # return with and without API flags
    valid_exts.append(base_ext)
    valid_exts = tuple(valid_exts)
    return valid_exts


def _autojit_cython(pyx_fpath, verbose=1, recompile=False, annotate=False):
    """
    This idea is that given a pyx file, we try to compile it. We write a stamp
    file so subsequent calls should be very fast as long as the source pyx has
    not changed.

    Parameters
    ----------
    pyx_fpath : str
        path to the pyx file

    verbose : int
        higher is more verbose.
    """
    import shutil
    if verbose > 3:
        print('_autojit_cython')

    # TODO: move necessary ubelt utilities to nx.utils?
    # Separate this into its own util?
    if shutil.which("cythonize"):
        pyx_dpath = dirname(pyx_fpath)

        if verbose > 3:
            print('pyx_dpath = {!r}'.format(pyx_dpath))

        # Check if the compiled library exists
        pyx_base = splitext(basename(pyx_fpath))[0]

        SO_EXTS = _platform_pylib_exts()
        so_fname = False
        for fname in os.listdir(pyx_dpath):
            if fname.startswith(pyx_base) and fname.endswith(SO_EXTS):
                so_fname = fname
                break

        if verbose > 3:
            print('so_fname = {!r}'.format(so_fname))

        try:
            # Currently this functionality depends on ubelt.
            # We could replace ub.cmd with subprocess.check_call and ub.augpath
            # with os.path operations, but hash_file and CacheStamp are harder
            # to replace. We can use "liberator" to statically extract these
            # and add them to nx.utils though.
            import ubelt as ub
        except Exception:
            if verbose > 3:
                print('return false, no ubelt')
            return False
        else:
            if so_fname is False:
                # We can compute what the so_fname will be if it doesnt exist
                so_fname = pyx_base + SO_EXTS[0]

            so_fpath = join(pyx_dpath, so_fname)
            content = ub.readfrom(pyx_fpath)
            mtime = os.stat(pyx_fpath).st_mtime

            depends = [ub.hash_data(content, hasher="sha1"), mtime]
            stamp_fname = ub.augpath(so_fname, ext=".jit.stamp")
            stamp = ub.CacheStamp(
                stamp_fname,
                dpath=pyx_dpath,
                product=so_fpath,
                depends=depends,
                verbose=verbose,
            )
            if verbose > 3:
                print('stamp = {!r}'.format(stamp))
            if recompile or stamp.expired():
                # Heuristic to try and grab the numpy include dir or not
                cythonize_args = ['cythonize']
                cythonize_env = os.environ.copy()
                needs_numpy = 'numpy' in content
                if needs_numpy:
                    import numpy as np
                    import pathlib
                    numpy_include_dpath = pathlib.Path(np.get_include())
                    numpy_dpath = (numpy_include_dpath / '../..').resolve()
                    # cythonize_env['CPATH'] = numpy_include_dpath + ':' + cythonize_env.get('CPATH', '')
                    cythonize_env['CFLAGS'] = ' '.join([
                        '-I{}'.format(numpy_include_dpath),
                    ]) + cythonize_env.get('CFLAGS', '')

                    cythonize_env['LDFLAGS'] = ' '.join([
                        '-L{} -lnpyrandom'.format(numpy_dpath / 'random/lib'),
                        '-L{} -lnpymath'.format(numpy_dpath / 'core/lib'),
                    ]) + cythonize_env.get('LDFLAGS', '')
                if annotate:
                    cythonize_args.append('-a')
                cythonize_args.append('-i {}'.format(pyx_fpath))
                cythonize_cmd = ' '.join(cythonize_args)
                if needs_numpy:
                    print('CFLAGS="{}" '.format(cythonize_env['CFLAGS']) + 'LDFLAGS="{}" '.format(cythonize_env['LDFLAGS']) + cythonize_cmd)
                ub.cmd(cythonize_cmd, verbose=verbose, check=True,
                       env=cythonize_env)
                stamp.renew()
            return True
    else:
        if verbose > 2:
            print('Cythonize not found!')



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
import scriptconfig as scfg
import argparse
import ubelt as ub
import sys


class RawDescriptionDefaultsHelpFormatter(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter):
    pass


_SUB_CLIS = []


def _register(cli_cls):
    _SUB_CLIS.append(cli_cls)
    return cli_cls


@_register
class CodeblockCLI(scfg.Config):
    name = 'codeblock'
    description = 'Remove indentation from text'
    default = {
        'text': scfg.Value('', position=1, help='text to dedent'),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        config = cls(cmdline=cmdline, data=kwargs)
        print(ub.codeblock(config['text']))


@_register
class InfoCLI(scfg.Config):
    name = 'info'
    description = 'Info about xdev'
    default = {}

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        import xdev
        print('sys.version_info = {!r}'.format(sys.version_info))
        print('xdev.__version__ = {!r}'.format(xdev.__version__))
        print('xdev.__file__ = {!r}'.format(xdev.__file__))


@_register
class DocstrStubgenCLI(scfg.Config):
    name = 'doctypes'
    description = 'Generate Typed Stubs from Docstrings'
    default = {
        'module': scfg.Value(None, position=1, help=ub.paragraph(
            '''
            The name of a module in the PYTHONPATH or an explicit path to that
            module.
            ''')),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        from xdev.cli import docstr_stubgen
        import ubelt as ub
        config = cls(cmdline=cmdline, data=kwargs)
        print(f'config={config}')
        modname_or_path = config['module']
        print(f'modname_or_path={modname_or_path}')
        if modname_or_path is None:
            raise ValueError('Must specify the module')
        modpath = docstr_stubgen.modpath_coerce(modname_or_path)
        modpath = ub.Path(modpath)
        generated = docstr_stubgen.generate_typed_stubs(modpath)

        for fpath, text in generated.items():
            fpath = ub.Path(fpath)
            print(f'Write fpath={fpath}')
            fpath.write_text(text)

        # Generate a py.typed file to mark the package as typed
        if modpath.is_dir():
            pytyped_fpath = (modpath / 'py.typed')
            print(f'touch pytyped_fpath={pytyped_fpath}')
            pytyped_fpath.touch()


@_register
class SedCLI(scfg.Config):
    name = 'sed'
    description = 'Search and replace text in files'
    default = {
        'regexpr': scfg.Value('', position=1, help=ub.paragraph(
            '''
            The pattern to search for.
            ''')),
        'repl': scfg.Value('', position=2, help=ub.paragraph(
            '''
            The pattern to replace with.
            ''')),
        'dpath': scfg.Value(None, position=3, help=ub.paragraph(
            '''
            The directory to recursively search or a file pattern to match.
            '''
        )),
        'dry': scfg.Value('ask', position=4, help=ub.paragraph(
            '''
            if 1, show what would be done. if 0, execute the change, if "ask",
            then show the dry run and then ask for confirmation.
            '''
        )),
        'include': scfg.Value(None),
        'exclude': scfg.Value(None),
        'recursive': scfg.Value(True),
        'verbose': scfg.Value(1),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        from xdev import search_replace
        config = cls(cmdline=cmdline, data=kwargs)
        if config['verbose'] > 2:
            print('config = {}'.format(ub.repr2(dict(config), nl=1, sort=0)))

        if config['dry'] in {'ask', 'auto'}:
            from rich.prompt import Confirm
            config['dry'] = True
            search_replace.sed(**config)
            flag = Confirm.ask('Do you want to execute this sed?')
            if flag:
                config['dry'] = False
                search_replace.sed(**config)
        else:
            search_replace.sed(**config)


@_register
class FormatQuotesCLI(scfg.Config):
    name = 'format_quotes'
    description = 'Use single quotes for code and double quotes for docs'
    default = {
        'path': scfg.Value('', position=1, help=ub.paragraph(
            '''
            ''')),
        'diff': scfg.Value(True, help=ub.paragraph(
            '''
            The pattern to replace with.
            ''')),
        'write': scfg.Value(False, short_alias=['-w'], help=ub.paragraph(
            '''
            The directory to recursively search or a file pattern to match.
            '''
        )),
        'verbose': scfg.Value(3, help=ub.paragraph(
            '''
            '''
        )),
        'recursive': scfg.Value(True),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        from xdev import format_quotes
        config = cls(cmdline=cmdline, data=kwargs)
        format_quotes.format_quotes(**config)


@_register
class FindCLI(scfg.Config):
    name = 'find'
    description = 'Find files based on names'
    default = {
        'pattern': scfg.Value('', position=1),
        'dpath': scfg.Value(None, position=2),
        'include': scfg.Value(None),
        'exclude': scfg.Value(None),
        'type': scfg.Value('f', help="can be f and/or d"),
        'recursive': scfg.Value(True),
        'followlinks': scfg.Value(False),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        from xdev import search_replace
        config = cls(cmdline=cmdline, data=kwargs)
        for found in search_replace.find(**config):
            print(found)


@_register
class TreeCLI(scfg.Config):
    name = 'tree'
    description = 'List a directory like a tree'

    default = {
        'cwd': scfg.Value('.', position=1),
        'max_files': scfg.Value(100),
        'colors': scfg.Value(not ub.NO_COLOR, isflag=True),
        'dirblocklist': scfg.Value(None)
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        import xdev
        config = cls(cmdline=cmdline, data=kwargs)
        print(xdev.tree_repr(**config))


@_register
class PintCLI(scfg.Config):
    """
    Converts one type of unit to another via the pint library.

    Notes:

    See Also
    --------
    The pint-convert tool comes pre-installed with pint but isn't as useful for
    in-bash computation unless you munge the output. The idea here is that when
    something like GDAL wants an environ in bytes, we can specify it in
    megabytes.

    Example Usage
    -------------
    xdev pint "10 megabytes" "bytes" --precision=0

    """
    __command__ = 'pint'
    description = 'Converts one type of unit to another via the pint library.'
    # input_expr = scfg.Value(None, position=1)
    # output_expr = scfg.Value(None, position=2)
    default = {
        'input_expr': scfg.Value(None, position=1),
        'output_unit': scfg.Value(None, position=2),
        'precision': scfg.Value(0, type=int),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        import pint
        ureg = pint.UnitRegistry()
        args = cls(cmdline=cmdline, data=kwargs)
        input = ureg.parse_expression(args['input_expr'])
        output_unit = args['output_unit']
        if output_unit is None:
            output_unit = input.unit
        output = input.to(output_unit)
        if args['precision'] == 0:
            print(int(output.magnitude))
        else:
            print(output.magnitude)


@_register
class ModpathCLI(scfg.Config):
    """
    Prints the path corresponding to a Python module

    Example Usage
    -------------
    # Simply ask where a module is stored
    xdev modpath xdev
    xdev modpath numpy

    # Use this feature in scripts for developement to avoid referencing
    # machine-specific paths.
    MODPATH=$(xdev modpath ubelt)
    echo "MODPATH = $MODPATH"
    """
    __command__ = 'modpath'
    description = 'Prints the path corresponding to a Python module'
    # input_expr = scfg.Value(None, position=1)
    # output_expr = scfg.Value(None, position=2)
    default = {
        'modname': scfg.Value(None, position=1),
    }

    @classmethod
    def main(cls, cmdline=False, **kwargs):
        args = cls(cmdline=cmdline, data=kwargs)
        modpath = ub.modname_to_modpath(args['modname'])
        print(modpath)


class ModalCLI(object):
    """
    Contains multiple scriptconfig.Config items with corresponding `main`
    functions.
    """
    def __init__(self, description='', sub_clis=[], version=None):
        self.description = description
        self.sub_clis = sub_clis
        self.version = version

    def build_parser(self):
        parser = argparse.ArgumentParser(
            description=self.description,
            formatter_class=RawDescriptionDefaultsHelpFormatter,
        )

        if self.version is not None:
            parser.add_argument('--version', action='store_true',
                                help='show version number and exit')

        subparsers = parser.add_subparsers(help='specify a command to run')
        for cli_cls in self.sub_clis:
            cmdname = getattr(cli_cls, '__command__', getattr(cli_cls, 'name', None))
            # cmdname = cli_cls.name
            subconfig = cli_cls()
            parserkw = {}
            aliases = getattr(cli_cls, 'aliases', [])
            if aliases:
                parserkw['aliases']  = aliases
            parserkw.update(subconfig._parserkw())
            parserkw['help'] = parserkw['description'].split('\n')[0]
            subparser = subparsers.add_parser(cmdname, **parserkw)
            subparser = subconfig.argparse(subparser)
            subparser.set_defaults(main=cli_cls.main)
        return parser

    def run(self):
        parser = self.build_parser()

        import os
        WATCH_LOOSE_CLI = os.environ.get('XDEV_LOOSE_CLI', '')
        if WATCH_LOOSE_CLI:
            ns = parser.parse_known_args()[0]
        else:
            ns = parser.parse_args()
        kw = ns.__dict__

        if kw.pop('version'):
            print(self.version)
            return 0

        sub_main = kw.pop('main', None)
        if sub_main is None:
            parser.print_help()
            raise ValueError('no command given')
            return 1

        try:
            ret = sub_main(cmdline=False, **kw)
        except Exception as ex:
            print('ERROR ex = {!r}'.format(ex))
            raise
            return 1
        else:
            if ret is None:
                ret = 0
            return ret


def main():
    import xdev
    modal_cli = ModalCLI(
        description='The XDEV CLI',
        version=xdev.__version__,
        sub_clis=_SUB_CLIS,
    )
    modal_cli.run()


if __name__ == '__main__':
    """
    CommandLine:
        xdev --help
        xdev --version
        xdev info
        xdev sed "main" "MAIN" "." --dry=True --include="*_*.py"
        xdev find "*_*.py"  '.'
        xdev codeblock "
            import sys
            print(sys.argv)
            print([
                'hello world'
            ])
        "
    """
    main()



import types
import ubelt as ub
from os.path import sys


def reload_class(self, verbose=True, reload_module=True):
    """
    Reload a class for a specific instance.

    (populates any new methods that may have just been written)

    Ported from utool
    """
    verbose = verbose
    classname = self.__class__.__name__
    try:
        modname = self.__class__.__module__
        if verbose:
            print('[class] reloading ' + classname + ' from ' + modname)
        # --HACK--
        if hasattr(self, '_on_reload'):
            if verbose > 1:
                print('[class] calling _on_reload for ' + classname)
            self._on_reload()
        elif verbose > 1:
            print('[class] ' + classname + ' does not have an _on_reload function')

        # Do for all inheriting classes
        def find_base_clases(_class, find_base_clases=None):
            class_list = []
            for _baseclass in _class.__bases__:
                parents = find_base_clases(_baseclass, find_base_clases)
                class_list.extend(parents)
            if _class is not object:
                class_list.append(_class)
            return class_list

        head_class = self.__class__
        # Determine if parents need reloading
        class_list = find_base_clases(head_class, find_base_clases)
        # HACK
        # ignore = {HashComparable2}
        ignore = {}
        class_list = [_class for _class in class_list
                      if _class not in ignore]
        for _class in class_list:
            if verbose:
                print('[class] reloading parent ' + _class.__name__ +
                      ' from ' + _class.__module__)
            if _class.__module__ == '__main__':
                # Attempt to find the module that is the main module
                # This may be very hacky and potentially break
                main_module_ = sys.modules[_class.__module__]
                main_modname = ub.modpath_to_modname(main_module_.__file__)
                module_ = sys.modules[main_modname]
            else:
                module_ = sys.modules[_class.__module__]
            if reload_module:
                import imp
                if verbose:
                    print('[class] reloading ' + _class.__module__ + ' with imp')
                try:
                    imp.reload(module_)
                except (ImportError, AttributeError):
                    print('[class] fallback reloading ' + _class.__module__ +
                          ' with imp')
                    # one last thing to try. probably used
                    # import_module_from_fpath when importing this module
                    imp.load_source(module_.__name__, module_.__file__)
            # Reset class attributes
            _newclass = getattr(module_, _class.__name__)
            _reload_class_methods(self, _newclass, verbose=verbose)

        # --HACK--
        # TODO: handle injected definitions
        if hasattr(self, '_initialize_self'):
            if verbose > 1:
                print('[class] calling _initialize_self for ' + classname)
            self._initialize_self()
        elif verbose > 1:
            print('[class] ' + classname + ' does not have an _initialize_self function')
    except Exception as ex:
        raise


def _reload_class_methods(self, class_, verbose=True):
    """
    rebinds all class methods

    Args:
        self (object): class instance to reload
        class_ (type): type to reload as
    """
    if verbose:
        print('[util_class] Reloading self=%r as class_=%r' % (self, class_))
    self.__class__ = class_
    for key in dir(class_):
        # Get unbound reloaded method
        func = getattr(class_, key)
        if isinstance(func, types.MethodType):
            # inject it into the old instance
            ub.inject_method(self, func)



import ubelt as ub
import numpy as np
from . import embeding
from . import util

__all__ = ['InteractiveIter']


INDEXABLE_TYPES = (list, tuple, np.ndarray)


class InteractiveIter(object):
    """
    Choose next value interactively

    iterable should be a list, not a generator. sorry

    CommandLine:
        xdoctest -m xdev.interactive_iter InteractiveIter:0 --interact
        xdoctest -m xdev.interactive_iter InteractiveIter:1 --interact

    Example:
        >>> # xdoctest: +REQUIRES(--interact)
        >>> from xdev.interactive_iter import *  # NOQA
        >>> iterable = [1, 2, 3]
        >>> enabled = True
        >>> startx = 0
        >>> default_action = 'next'
        >>> custom_actions = []
        >>> wraparound = False
        >>> display_item = True
        >>> verbose = True
        >>> iiter = InteractiveIter(iterable, enabled, startx, default_action, custom_actions, wraparound, display_item, verbose)
        >>> for _ in iiter:
        >>>     pass

    Example:
        >>> # xdoctest: +REQUIRES(--interact)
        >>> # Interactive matplotlib stuff
        >>> from xdev.interactive_iter import *  # NOQA
        >>> import kwimage
        >>> import kwplot
        >>> plt = kwplot.autoplt(verbose=3, force='Qt5Agg')
        >>> plt.ion()
        >>> keys = list(kwimage.grab_test_image.keys())
        >>> iterable = [kwimage.grab_test_image(key) for key in keys]
        >>> iiter = InteractiveIter(iterable)
        >>> for img in iiter:
        >>>     kwplot.imshow(img)
        >>>     InteractiveIter.draw()

    """
    def __init__(iiter, iterable=None, enabled=True, startx=0,
                 default_action='next', custom_actions=[], wraparound=False,
                 display_item=False, verbose=True):
        r"""
        Args:
            iterable (None): (default = None)
            enabled (bool): (default = True)
            startx (int): (default = 0)
            default_action (str): (default = 'next')
            custom_actions (list): list of 4-tuple (name, actions, help, func) (default = [])
            wraparound (bool): (default = False)
            display_item (bool): (default = True)
            verbose (bool):  verbosity flag(default = True)
        """
        iiter.wraparound = wraparound
        iiter.enabled = enabled
        iiter.iterable = iterable

        for actiontup in custom_actions:
            if isinstance(custom_actions, tuple):
                pass
            else:
                pass

        iiter.custom_actions = util.take_column(custom_actions, [0, 1, 2])
        iiter.custom_funcs = util.take_column(custom_actions, 3)
        iiter.action_tuples = [
            # (name, list, help)
            ('next',   ['n'], 'move to the next index'),
            ('prev',   ['p'], 'move to the previous index'),
            ('reload', ['r'], 'stay at the same index'),
            ('index',  ['x', 'i', 'index'], 'move to that index'),
            ('set',    ['set'], 'set current index value'),
            ('ipy',    ['ipy', 'ipython', 'cmd'], 'start IPython'),
            ('quit',   ['q', 'exit', 'quit'], 'quit'),
        ] + iiter.custom_actions
        default_action_index = util.take_column(iiter.action_tuples, 0).index(default_action)
        iiter.action_tuples[default_action_index][1].append('')
        iiter.action_keys = {tup[0]: tup[1] for tup in iiter.action_tuples}
        iiter.index = startx
        iiter.display_item = display_item
        iiter.verbose = verbose

    @classmethod
    def eventloop(cls, custom_actions=[]):
        """
        For use outside of iteration wrapping. Makes an interactive event loop
        custom_actions should be specified in format
        [dispname, keys, desc, func]
        """
        iiter = cls([None], custom_actions=custom_actions, verbose=False)
        print('[IITER] Begining interactive main loop')
        for _ in iiter:
            pass
        return iiter

    def __iter__(iiter):
        global LIVE_INTERACTIVE_ITER
        LIVE_INTERACTIVE_ITER = iiter
        if not iiter.enabled:
            for item in ub.ProgIter(iiter.iterable, desc='nointeract: ', freq=1, adjust=False):
                yield item
            return
            # raise StopIteration()
        assert isinstance(iiter.iterable, INDEXABLE_TYPES), 'input is not iterable'
        iiter.num_items = len(iiter.iterable)
        if iiter.verbose:
            print('[IITER] Begin interactive iteration: %r items\n' % (iiter.num_items))
        if iiter.num_items == 0:
            return
            # raise StopIteration
        # TODO: replace with ub.ProgIter
        # mark_, end_ = util_progress.log_progress(length=iiter.num_items,
        #                                          lbl='interaction: ', freq=1)
        prog = ub.ProgIter(total=iiter.num_items, desc='interaction: ', freq=1,
                           show_times=True, verbose=2)

        prog.begin()
        prompt_on_start = False
        if prompt_on_start:
            ans = iiter.prompt()
            action = iiter.handle_ans(ans)
            REFRESH_ON_BAD_INPUT = False
            if not REFRESH_ON_BAD_INPUT:
                while action is False:
                    ans = iiter.prompt()
                    action = iiter.handle_ans(ans)
            if action == 'IPython':
                embeding.embed(n=1)

        while True:
            if iiter.verbose:
                print('')
            if iiter.wraparound:
                iiter.index = iiter.index % len(iiter.iterable)
            if iiter.index >= len(iiter.iterable):
                if iiter.verbose:
                    print('Got to end the end of the iterable')
                break

            # Hack to jump to a particular index
            # (TODO: update ProgIter to handle this)
            # offset = iiter.index - prog._iter_idx
            # prog._reset_internals()
            # prog.step(iiter.index + 1)

            item = iiter.iterable[iiter.index]
            if iiter.verbose:
                print('')
            yield item
            if iiter.verbose:
                print('')

            # prog._reset_internals()
            # offset = iiter.index - prog._iter_idx
            # prog.step(offset)
            prog._reset_internals()
            if iiter.index > 0:
                prog.step(iiter.index)
            else:
                prog.display_message()

            if iiter.verbose:
                print('')
                print('[IITER] current index=%r' % (iiter.index,))
                if iiter.display_item:
                    print('[IITER] current item=%r' % (item,))
            ans = iiter.prompt()
            try:
                action = iiter.handle_ans(ans)
            except StopIteration:
                return
            REFRESH_ON_BAD_INPUT = False
            if not REFRESH_ON_BAD_INPUT:
                while action is False:
                    ans = iiter.prompt()
                    try:
                        action = iiter.handle_ans(ans)
                    except StopIteration:
                        return
            if action == 'IPython':
                embeding.embed(n=1)

        prog.end()
        print('Ended interactive iteration')
        LIVE_INTERACTIVE_ITER = None

    def handle_ans(iiter, ans_):
        """
        preforms an actionm based on a user answer
        """
        ans = ans_.strip(' ')
        def parse_str_value(ans):
            return ' '.join(ans.split(' ')[1:])
        def chack_if_answer_was(valid_keys):
            return any([ans == key or ans.startswith(key + ' ') for key in valid_keys])
        # Handle standard actions
        if ans in iiter.action_keys['quit']:
            raise StopIteration()
        elif ans in iiter.action_keys['prev']:
            iiter.index -= 1
        elif ans in iiter.action_keys['next']:
            iiter.index += 1
        elif ans in iiter.action_keys['reload']:
            iiter.index += 0
        elif chack_if_answer_was(iiter.action_keys['index']):
            try:
                iiter.index = int(parse_str_value(ans))
            except ValueError:
                print('Unknown ans=%r' % (ans,))
        elif chack_if_answer_was(iiter.action_keys['set']):
            try:
                iiter.iterable[iiter.index] = eval(parse_str_value(ans))
            except ValueError:
                print('Unknown ans=%r' % (ans,))
        elif ans in iiter.action_keys['ipy']:
            return 'IPython'
        else:
            # Custom interactions
            for func, tup in zip(iiter.custom_funcs, iiter.custom_actions):
                key = tup[0]
                if chack_if_answer_was(iiter.action_keys[key]):
                    value  = parse_str_value(ans)
                    # cal custom function
                    print('Calling custom action func')
                    import inspect
                    argspec = inspect.getfullargspec(func)
                    if len(argspec.args) == 3:
                        # Forgot why I had custom functions take args in the first place
                        func(iiter, key, value)
                    else:
                        func()
                    # Custom funcs dont cause iteration
                    return False
            print('Unknown ans=%r' % (ans,))
            return False
        return True

    def prompt(iiter):
        def _or_phrase(list_):
            return util.conj_phrase(list(map(repr, map(str, list_))), 'or')
        msg_list = ['enter %s to %s' % (_or_phrase(tup[1]), tup[2])
                    for tup in iiter.action_tuples]
        msg = ub.indent('\n'.join(msg_list), ' | * ')
        msg = ''.join([' +-----------', msg, '\n L-----------\n'])
        # TODO: timeout, help message
        print(msg)
        ans = iiter.wait_for_input()
        return ans

    def wait_for_input(iiter):
        ans = input().strip()
        return ans

    def __call__(iiter, iterable=None):
        iiter.iterable = iterable

    @classmethod
    def draw(iiter):
        """
        in the common case where InteractiveIter is used to view matplotlib
        figures, you will have to draw the figure manually. This is a helper
        for that task.
        """
        from matplotlib import pyplot as plt
        fig = plt.gcf()
        fig.canvas.draw()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub


def quantum_random(pure=False):
    """
    Returns a quantum random number as a 32 bit unsigned integer.
    Does this by making a network request to the ANU Quantum Random Number
    Generator web service, so an internet connection is required.

    Args:
        pure (bool): if False, mixes this data with pseudorandom data for
            security. Otherwise returns the raw quantum numbers that were
            sent over the web (i.e. subject to MitM attacks).

    Requirements:
        quantumrandom >= 1.9.0

    Returns:
        numpy.uint32: the random number
    """
    import numpy as np
    import os
    import quantumrandom

    # Data was sent over a network
    qr_data16 = quantumrandom.uint16(array_length=2)
    nbytes = qr_data16.size * qr_data16.dtype.itemsize

    if pure:
        data16 = qr_data16
    else:
        # Cryptographically generated
        buf = memoryview(os.urandom(nbytes))
        pr_data16 = np.frombuffer(buf, dtype=qr_data16.dtype)
        # xor to mix data
        data16 = (pr_data16 ^ qr_data16)

    assert data16.flags['C_CONTIGUOUS']
    data32 = data16.view(np.dtype('uint32'))[0]
    return data32


def byte_str(num, unit='auto', precision=2):
    """
    Automatically chooses relevant unit (KB, MB, or GB) for displaying some
    number of bytes.

    Args:
        num (int): number of bytes
        unit (str): which unit to use, can be auto, B, KB, MB, GB, TB, PB, EB,
            ZB, or YB.
        precision (int): number of decimals of precision

    References:
        https://en.wikipedia.org/wiki/Orders_of_magnitude_(data)

    Returns:
        str: string representing the number of bytes with appropriate units

    Example:
        >>> num_list = [1, 100, 1024,  1048576, 1073741824, 1099511627776]
        >>> result = ub.repr2(list(map(byte_str, num_list)), nl=0)
        >>> print(result)
        ['0.00 KB', '0.10 KB', '1.00 KB', '1.00 MB', '1.00 GB', '1.00 TB']
    """
    abs_num = abs(num)
    if unit == 'auto':
        if abs_num < 2.0 ** 10:
            unit = 'KB'
        elif abs_num < 2.0 ** 20:
            unit = 'KB'
        elif abs_num < 2.0 ** 30:
            unit = 'MB'
        elif abs_num < 2.0 ** 40:
            unit = 'GB'
        elif abs_num < 2.0 ** 50:
            unit = 'TB'
        elif abs_num < 2.0 ** 60:
            unit = 'PB'
        elif abs_num < 2.0 ** 70:
            unit = 'EB'
        elif abs_num < 2.0 ** 80:
            unit = 'ZB'
        else:
            unit = 'YB'
    if unit.lower().startswith('b'):
        num_unit = num
    elif unit.lower().startswith('k'):
        num_unit =  num / (2.0 ** 10)
    elif unit.lower().startswith('m'):
        num_unit =  num / (2.0 ** 20)
    elif unit.lower().startswith('g'):
        num_unit = num / (2.0 ** 30)
    elif unit.lower().startswith('t'):
        num_unit = num / (2.0 ** 40)
    elif unit.lower().startswith('p'):
        num_unit = num / (2.0 ** 50)
    elif unit.lower().startswith('e'):
        num_unit = num / (2.0 ** 60)
    elif unit.lower().startswith('z'):
        num_unit = num / (2.0 ** 70)
    elif unit.lower().startswith('y'):
        num_unit = num / (2.0 ** 80)
    else:
        raise ValueError('unknown num={!r} unit={!r}'.format(num, unit))
    return ub.repr2(num_unit, precision=precision) + ' ' + unit


def set_overlaps(set1, set2, s1='s1', s2='s2'):
    """
    return info about set overlaps

    Notes:
        This function needs a rename. Possible candidates brainstorm:
            * set_analysis
            * set_binary_analysis
            * set_binary_describe
            * set_relationships
            * describe_sets
            * describe_relations
            * describe_set_relations
            * sets_summary
            * sumarize_sets
            * sumerset
    """
    set1 = set(set1)
    set2 = set(set2)
    overlaps = ub.odict([
        (s1, len(set1)),
        (s2, len(set2)),
        ('isect', len(set1.intersection(set2))),
        ('union', len(set1.union(set2))),
        ('%s - %s' % (s1, s2), len(set1.difference(set2))),
        ('%s - %s' % (s2, s1), len(set2.difference(set1))),
    ])
    return overlaps


def nested_type(obj, unions=False):
    """
    Compute the :module:`typing` compatible annotation type.

    Args:
        obj (Any): a typing template based on a specific object
        unions (bool): if True use unions, otherwise use Any

    Returns:
        str: type code (might change to return actual type)

    Example:
        >>> obj = {'a': [1, 2], 'b': [3, 4, 5]}
        >>> print(nested_type(obj))
        Dict[str, List[int]]

        >>> import numpy as np
        >>> obj = {'b': {'a': 1.0, 'b': 'foo', 'c': np.array([1, 2])}}
        >>> print(nested_type(obj, unions=True))
        Dict[str, Dict[str, float | ndarray | str]]
    """
    def _resolve(_types):
        if len(_types) == 1:
            return ub.peek(_types)
        else:
            if unions:
                return ' | '.join(sorted(_types))
            else:
                return 'Any'

    from functools  import partial
    _nested = partial(nested_type, unions=unions)
    if isinstance(obj, dict):
        keytypes = {_nested(k) for k in obj.keys()}
        valtypes = {_nested(v) for v in obj.values()}
        keytype = _resolve(keytypes)
        valtype = _resolve(valtypes)
        objtype = 'Dict[{}, {}]'.format(keytype, valtype)
    elif isinstance(obj, list):
        itemtypes = {_nested(item) for item in obj}
        itemtype = _resolve(itemtypes)
        objtype = 'List[{}]'.format(itemtype)
    elif isinstance(obj, set):
        itemtypes = [_nested(item) for item in obj]
        itemtype = _resolve(itemtypes)
        objtype = 'Set[{}]'.format(itemtype)
    elif isinstance(obj, tuple):
        itemtypes = [_nested(item) for item in obj]
        objtype = 'Tuple[{}]'.format(', '.join(itemtypes))
    else:
        import typing
        objtype = type(obj).__name__
        if hasattr(typing, '_normalize_alias'):
            objtype = typing._normalize_alias.get(objtype, objtype)
        else:
            objtype = {'list': 'List',
                       'tuple': 'Tuple',
                       'dict': 'Dict',
                       'set': 'Set',
                       'frozenset': 'FrozenSet',
                       'deque': 'Deque',
                       'defaultdict': 'DefaultDict',
                       'type': 'Type',
                       'Set': 'AbstractSet'}.get(objtype, objtype)
        return objtype
    return objtype


def difftext(text1, text2, context_lines=0, ignore_whitespace=False,
             colored=False):
    r"""
    Uses difflib to return a difference string between two similar texts

    Args:
        text1 (str): old text
        text2 (str): new text
        context_lines (int): number of lines of unchanged context
        ignore_whitespace (bool):
        colored (bool): if true highlight the diff

    Returns:
        str: formatted difference text message

    References:
        http://www.java2s.com/Code/Python/Utility/IntelligentdiffbetweentextfilesTimPeters.htm

    Example:
        >>> # build test data
        >>> text1 = 'one\ntwo\nthree'
        >>> text2 = 'one\ntwo\nfive'
        >>> # execute function
        >>> result = difftext(text1, text2)
        >>> # verify results
        >>> print(result)
        - three
        + five

    Example:
        >>> # build test data
        >>> text1 = 'one\ntwo\nthree\n3.1\n3.14\n3.1415\npi\n3.4\n3.5\n4'
        >>> text2 = 'one\ntwo\nfive\n3.1\n3.14\n3.1415\npi\n3.4\n4'
        >>> # execute function
        >>> context_lines = 1
        >>> result = difftext(text1, text2, context_lines, colored=True)
        >>> # verify results
        >>> print(result)
    """
    import ubelt as ub
    import difflib
    text1 = ub.ensure_unicode(text1)
    text2 = ub.ensure_unicode(text2)
    text1_lines = text1.splitlines()
    text2_lines = text2.splitlines()
    if ignore_whitespace:
        text1_lines = [t.rstrip() for t in text1_lines]
        text2_lines = [t.rstrip() for t in text2_lines]
        ndiff_kw = dict(linejunk=difflib.IS_LINE_JUNK,
                        charjunk=difflib.IS_CHARACTER_JUNK)
    else:
        ndiff_kw = {}
    all_diff_lines = list(difflib.ndiff(text1_lines, text2_lines, **ndiff_kw))

    if context_lines is None:
        diff_lines = all_diff_lines
    else:
        # boolean for every line if it is marked or not
        ismarked_list = [len(line) > 0 and line[0] in '+-?'
                         for line in all_diff_lines]
        # flag lines that are within context_lines away from a diff line
        isvalid_list = ismarked_list[:]
        for i in range(1, context_lines + 1):
            isvalid_list[:-i] = list(map(any, zip(
                isvalid_list[:-i], ismarked_list[i:])))
            isvalid_list[i:] = list(map(any, zip(
                isvalid_list[i:], ismarked_list[:-i])))

        USE_BREAK_LINE = True
        if USE_BREAK_LINE:
            # insert a visual break when there is a break in context
            diff_lines = []
            prev = False
            visual_break = '\n <... FILTERED CONTEXT ...> \n'
            #print(isvalid_list)
            for line, valid in zip(all_diff_lines, isvalid_list):
                if valid:
                    diff_lines.append(line)
                elif prev:
                    if False:
                        diff_lines.append(visual_break)
                prev = valid
        else:
            diff_lines = list(ub.compress(all_diff_lines, isvalid_list))
    text = '\n'.join(diff_lines)
    if colored:
        text = ub.highlight_code(text, lexer_name='diff')
    return text


def tree_repr(cwd=None, max_files=100, dirblocklist=None, show_nfiles='auto',
              return_text=False, return_tree=False, pathstyle='name',
              with_type=False, colors=not ub.NO_COLOR):
    """
    Filesystem tree representation

    Like the unix util tree, but allow writing numbers of files per directory
    when given -d option

    cwd = '/data/public/Aerial/US_ALASKA_MML_SEALION'

    Args:
        cwd : directory to print
        max_files : maximum files to print before supressing a directory
        pathstyle (str): can be rel, name, or abs
        return_tree (bool): if True return the tree
        return_text (bool): if True return the text
        colors (bool): if True use rich

    SeeAlso:
        xdev.tree - generator

    Ignore:
        >>> import xdev
        >>> xdev.tree_repr()
    """
    import os
    from os.path import join, relpath, basename
    if cwd is None:
        cwd = os.getcwd()

    import networkx as nx
    tree = nx.OrderedDiGraph()

    from xdev.patterns import MultiPattern
    if dirblocklist is not None:
        dirblocklist = MultiPattern.coerce(dirblocklist, hint='glob')

    def _make_label(p):
        if pathstyle == 'rel':
            pathrep = relpath(p, cwd)
        elif pathstyle == 'name':
            pathrep = basename(p)
        elif pathstyle == 'abs':
            pathrep = p
        else:
            KeyError(pathstyle)

        types = []
        islink = os.path.islink(p)
        isdir = os.path.isdir(p)
        isfile = os.path.isfile(p)
        isbroken = False
        scolor = ''
        tcolor = ''
        L_scolor = ''
        L_tcolor = ''
        if islink:
            if colors:
                L_scolor = '[cyan]'
                L_tcolor = '[/cyan]'
            types.append('L')
            if not isfile and not isdir:
                isbroken = True
                if isbroken:
                    if colors:
                        L_scolor = '[red]'
                        L_tcolor = '[/red]'
                types.append('B')

        if isfile:
            if colors:
                scolor = '[reset]'
                tcolor = '[/reset]'
            types.append('F')
        if isdir:
            if colors:
                scolor = '[blue]'
                tcolor = '[/blue]'
            types.append('D')

        if islink:
            target = os.readlink(p)
            pathrep = L_scolor + pathrep + L_tcolor + ' -> ' + scolor + target + tcolor
        else:
            pathrep = scolor + pathrep + tcolor

        if with_type:
            typelabel = ''.join(types)
            return f'({typelabel}) ' + pathrep
        else:
            return pathrep

    # TODO: rectify with "find"
    for root, dnames, fnames in os.walk(cwd):

        if dirblocklist is not None:
            dnames[:] = [
                dname for dname in dnames if not dirblocklist.match(dname)]

        dnames[:] = sorted(dnames)
        tree.add_node(root)

        too_many_files = max_files is not None and len(fnames) >= max_files

        if show_nfiles == 'auto':
            show_nfiles_ = too_many_files
        else:
            show_nfiles_ = show_nfiles

        num_files = len(fnames)
        if show_nfiles_:
            prefix = '[ {} ] '.format(num_files)
        else:
            prefix = ''

        label = '{}{}'.format(prefix, _make_label(root))

        tree.nodes[root]['label'] = label

        if not too_many_files:
            for fname in fnames:
                fpath = join(root, fname)
                tree.add_node(fpath)
                tree.nodes[fpath]['label'] = _make_label(fpath)
                tree.add_edge(root, fpath)

        for dname in dnames:
            dpath = join(root, dname)
            tree.add_node(dpath)
            tree.nodes[dpath]['label'] = _make_label(dpath)
            tree.add_edge(root, dpath)

    from xdev.util_networkx import write_network_text
    import io
    file = io.StringIO()
    write_network_text(tree, file)
    file.seek(0)
    text = file.read()

    info = {}

    if return_text:
        info['text'] = text
    else:
        if colors:
            from rich import print as rprint
            rprint(text)
        else:
            print(text)

    if return_tree:
        info['tree'] = tree
    return info




def make_warnings_print_tracebacks():
    import warnings
    import traceback
    _orig_formatwarning = warnings.formatwarning
    warnings._orig_formatwarning = _orig_formatwarning
    def _monkeypatch_formatwarning_tb(*args, **kwargs):
        s = _orig_formatwarning(*args, **kwargs)
        if len(s.strip()):
            tb = traceback.format_stack()
            s += ''.join(tb[:-1])
        return s
    warnings.formatwarning = _monkeypatch_formatwarning_tb



"""
Python implementations of sed, grep, and find

Porting from ~/local/rob/rob/rob_nav.py / ubelt
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import ubelt as ub
from os.path import relpath, split, join, abspath
from xdev.patterns import Pattern, RE_Pattern  # NOQA
from xdev.patterns import MultiPattern

# try:
#     from packaging.version import parse as parse_version
# except Exception:
#     from distutils.version import LooseVersion as parse_version


def sed(regexpr, repl, dpath=None, include=None, exclude=None, recursive=True,
        dry=False, verbose=1):
    r"""
    Execute a sed on multiple files.

    Example:
        >>> from xdev.search_replace import *  # NOQA
        >>> from xdev.search_replace import _create_test_filesystem
        >>> dpath = _create_test_filesystem()['root']
        >>> sed('a', 'x', dpath=dpath, dry=True)
    """
    num_changed = 0
    num_files_checked = 0
    num_skipped = 0
    fpaths_changed = []

    fpath_generator = find(dpath=dpath, type='f', include=include,
                           exclude=exclude, recursive=recursive)
    for fpath in fpath_generator:
        try:
            changed_lines = sedfile(fpath, regexpr, repl, dry=dry)
        except UnicodeDecodeError:
            num_skipped += 1
        else:
            num_files_checked += 1
            if len(changed_lines) > 0:
                fpaths_changed.append(fpath)
                num_changed += len(changed_lines)

    if verbose:
        print('num_files_checked = {}'.format(num_files_checked))
        print('num probable binary files skipped = {}'.format(num_skipped))
        print('fpaths_changed = {}'.format(ub.repr2(sorted(fpaths_changed))))
        print('total lines changed = {!r}'.format(num_changed))


def grep(regexpr, dpath=None, include=None, exclude=None, recursive=True,
         verbose=1):
    r"""
    Execute a grep on multiple files.

    Example:
        >>> from xdev.search_replace import *  # NOQA
        >>> from xdev.search_replace import _create_test_filesystem
        >>> dpath = _create_test_filesystem()['root']
        >>> grep('a', dpath=dpath)
    """
    grep_results = []

    fpath_generator = find(dpath=dpath, type='f', include=include,
                           exclude=exclude, recursive=recursive)

    for fpath in fpath_generator:
        grepres = grepfile(fpath, regexpr, verbose=verbose)
        if grepres:
            grep_results.append(grepres)

    if verbose:
        print('====================')
        print('====================')
        found_fpaths = [r.fpath for r in grep_results]
        print('\n'.join(found_fpaths))

    return grep_results


def find(pattern=None, dpath=None, include=None, exclude=None, dirblocklist=None,
         type=None, recursive=True, followlinks=False):
    """
    Find all paths in a root subject to a search criterion

    Args:
        pattern (str | None):
            The glob pattern the path name must match to be returned

        dpath (str | None):
            The root direcotry to search. Default to cwd.

        include (str | List[str] | MultiPattern | None):
            Pattern or list of patterns. If specified, search only files whose
            base name matches this pattern. By default the pattern is GLOB.
            This only applies to the final name. Directories that do not match
            this name will still be traversed.

        exclude (str | List[str] | MultiPattern | None):
            Pattern or list of patterns. Skip any file with a name suffix that
            matches the pattern. By default the pattern is GLOB. This ONLY
            applies to the final name. Directories that match an exclude
            pattern will still be traversed.  Use ``dirblocklist`` to specify
            patterns to exclude intermediate directories from traversal.

        dirblocklist (str | List[str] | MultiPattern | None):
            Any directory name matching this pattern will be removed from
            traversal.

        type (str | List[str] | None):
            A list of 1 character codes indicating what types of file can be
            returned. Currently we only allow either "f" for file or "d" for
            directory. Symbolic links are not currently distinguished. In the
            future we may support posix codes, see [1]_ for details.

        recursive:
            search all subdirectories recursively

        followlinks (bool, default=False):
            if True will follow directory symlinks

    References:
        _[1] https://linuxconfig.org/identifying-file-types-in-linux

    TODO:
        mindepth

        maxdepth

        ignore_case

        regex_match


    Example:
        >>> from xdev.search_replace import *  # NOQA
        >>> from xdev.search_replace import _create_test_filesystem
        >>> dpath = _create_test_filesystem()['root']
        >>> paths = list(find(pattern='*', dpath=dpath))
        >>> assert len(paths) == 5
        >>> paths = list(find(pattern='*', dpath=dpath, type='f'))
        >>> assert len(paths) == 4
    """

    if pattern is None:
        pattern = '*'

    if type is None:
        with_dirs = True
        with_files = True
    else:
        with_dirs = False
        with_files = False
        if 'd' in type:
            with_dirs = True
        if 'f' in type:
            with_files = True

    if dpath is None:
        dpath = os.getcwd()

    # Define helper for checking inclusion / exclusion
    include = None if include is None else MultiPattern.coerce(include, hint='glob')
    exclude = None if exclude is None else MultiPattern.coerce(exclude, hint='glob')
    dirblocklist = None if dirblocklist is None else MultiPattern.coerce(dirblocklist, hint='glob')
    main_pattern = Pattern.coerce(pattern, hint='glob')

    def is_included(name):
        if not main_pattern.match(name):
            return False

        if exclude is not None and exclude.match(name):
            return False

        if include is None or include.match(name):
            return True

        return False

    for root, dnames, fnames in os.walk(dpath, followlinks=followlinks):

        if dirblocklist is not None:
            dnames[:] = [
                dname for dname in dnames if not dirblocklist.match(dname)]

        if with_files:
            for fname in fnames:
                if is_included(fname):
                    yield join(root, fname)

        if with_dirs:
            for dname in dnames:
                if is_included(dname):
                    yield join(root, dname)

        if not recursive:
            break


def sedfile(fpath, regexpr, repl, dry=False, verbose=1):
    r"""
    Execute a search and replace on a particular file

    TODO:
        - [ ] Store "SedResult" class, with lazy execution

    Example:
        >>> from xdev.search_replace import *  # NOQA
        >>> from xdev.search_replace import _create_test_filesystem
        >>> fpath = _create_test_filesystem()['contents'][1]
        >>> changed_lines1 = sedfile(fpath, 'a', 'x', dry=True, verbose=1)
        >>> changed_lines2 = sedfile(fpath, 'a', 'x', dry=False, verbose=0)
        >>> assert changed_lines2 == changed_lines1
        >>> changed_lines3 = sedfile(fpath, 'a', 'x', dry=False, verbose=0)
        >>> assert changed_lines3 != changed_lines2
    """
    import xdev
    mode_text = ['(real-run)', '(dry-run)'][dry]

    pattern = Pattern.coerce(regexpr, hint='regex')
    # print(f'regexpr={regexpr}')
    # print(f'pattern={pattern!r}')

    path, name = split(fpath)
    new_file_lines = []
    try:
        with open(fpath, 'r') as file:
            file_lines = file.readlines()
            # Search each line for the desired regexpr
            new_file_lines = [pattern.sub(repl, line) for line in file_lines]
    except UnicodeDecodeError as ex:
        # Add the file name into the exception
        new_last_arg = ex.args[-1] + ' in fpath={!r}'.format(fpath)
        new_args = ex.args[:-1] + (new_last_arg,)
        raise UnicodeDecodeError(*new_args) from ex
    except Exception:
        raise
        # This does not preserve exception type
        # raise Exception('Failed to sedfile fpath = {!r}'.format(fpath)) from ex

    changed_lines = [(newline, line)
                     for newline, line in zip(new_file_lines, file_lines)
                     if  newline != line]
    nChanged = len(changed_lines)
    if nChanged > 0:
        try:
            rel_fpath = relpath(fpath, os.getcwd())
        except ValueError:
            # windows issues
            rel_fpath = abspath(fpath)

        if verbose:
            print(' * {} changed {} lines in {!r} '.format(
                mode_text, nChanged, rel_fpath))
            print(' * --------------------')
        new_file = ''.join(new_file_lines)
        old_file = ub.ensure_unicode(
            ''.join(list(map(ub.ensure_unicode, file_lines))))
        if verbose:
            print(xdev.difftext(old_file, new_file, colored=True))
        if not dry:
            if verbose:
                print(' ! WRITING CHANGES')
            with open(fpath, 'w') as file:
                file.write(new_file)
        return changed_lines
    return []


class GrepResult(ub.NiceRepr):
    """
    Manage and format results from grep
    """
    def __init__(self, fpath, pattern=None):
        self.pattern = pattern
        self.fpath = fpath
        self.found_lxs = []
        self.found_lines = []
        self.max_line = 100

    def __nice__(self):
        return '{} in {}'.format(len(self), self.fpath)

    def __iter__(self):
        return iter(self.found_lines)

    def __len__(self):
        return len(self.found_lines)

    def append(self, lx, line):
        self.found_lines.append(line)
        self.found_lxs.append(lx)

    def format_text(self, color=True):
        summary = []
        app = summary.append
        fname = ub.Path(self.fpath).name
        ndigits = str(len(str(self.max_line)))

        fmt_str = '{} : {:' + ndigits + 'd} |{}'
        ret = 'Found {} line(s) in {!r}: '.format(len(self), self.fpath)
        app('----------------------')
        color = 'red'
        app(ret)
        for (lx, line) in zip(self.found_lxs, self.found_lines):
            line = line.replace('\n', '')
            if color and self.pattern:
                found = self.pattern.search(line)
                s, t = found.span()
                line = line[:s] + ub.color_text(line[s:t], color) + line[t:]
            app(fmt_str.format(fname, lx, line))

        return '\n'.join(summary)


def grepfile(fpath, regexpr, verbose=1):
    r"""
    Exceute grep on a single file

    Example:
        >>> from xdev.search_replace import *  # NOQA
        >>> from xdev.search_replace import _create_test_filesystem
        >>> fpath = _create_test_filesystem()['contents'][1]
        >>> grep_result = grepfile(fpath, r'\bb\b')
        >>> print('grep_result = {}'.format(grep_result))
    """
    grep_result = None
    pattern = Pattern.coerce(regexpr, hint='regex')
    with open(fpath, 'r') as file:
        try:
            lines = file.readlines()
        except UnicodeDecodeError:
            print("UNABLE TO READ fpath={}".format(fpath))
        else:
            grep_result = GrepResult(fpath, pattern)
            grep_result.max_line = len(lines)

            # Search each line for the desired pattern
            for lx, line in enumerate(lines):
                match_object = pattern.search(line)
                if match_object:
                    grep_result.append(lx, line)

            # Print the results (if any)
            if verbose:
                if len(grep_result):
                    print(grep_result.format_text())

    return grep_result


def greptext(text, regexpr, fpath=None, verbose=1):
    r"""
    Exceute grep on text

    TODO: get context
    """
    from xdev.patterns import Pattern
    # from xdev.search_replace import GrepResult
    grep_result = None
    pattern = Pattern.coerce(regexpr, hint='regex')
    fpath = '<text>'
    try:
        lines = text.splitlines()
    except UnicodeDecodeError:
        print("UNABLE TO READ fpath={}".format(fpath))
    else:
        grep_result = GrepResult(fpath, pattern)
        grep_result.max_line = len(lines)

        # Search each line for the desired pattern
        for lx, line in enumerate(lines):
            match_object = pattern.search(line)
            if match_object:
                grep_result.append(lx, line)

        # Print the results (if any)
        if verbose:
            if len(grep_result):
                print(grep_result.format_text())
    return grep_result


def _create_test_filesystem():
    dpath = ub.ensure_app_cache_dir('xdev/test_search_replace')
    text1 = ub.paragraph(
        '''
        Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
        tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim
        veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea
        commodo consequat. Duis aute irure dolor in reprehenderit in voluptate
        velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint
        occaecat cupidatat non proident, sunt in culpa qui officia deserunt
        mollit anim id est laborum.
        ''')

    text2 = ub.codeblock(
        '''
        def fib(n):
            a, b = 0, 1
            while a < n:
                print(a, end=' ')
                a, b = b, a+b
                print()

        fib(1000)
        ''')

    text3 = ub.codeblock(
        '''
        This file contains Lorem and fib

        Newlines

        fib

        lorem

        fib
        ''')

    text4 = ''

    fpath1 = join(dpath, 'lorium.txt')
    fpath2 = join(dpath, 'fib.py')
    fpath3 = join(dpath, 'foo.txt')
    fpath4 = join(ub.ensuredir((dpath, 'subdir')), 'foo.txt')
    with open(fpath1, 'w') as file:
        file.write(text1)
    with open(fpath2, 'w') as file:
        file.write(text2)
    with open(fpath3, 'w') as file:
        file.write(text3)
    with open(fpath4, 'w') as file:
        file.write(text4)

    info = {
        'root': dpath,
        'contents': [fpath1, fpath2, fpath3],
    }

    return info

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/xdev/xdev/search_replace.py
        xdoctest ~/code/xdev/xdev/search_replace.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
